# -*- coding: utf-8 -*-
"""
author: zengbin93
email: zeng_bin8888@163.com
create_dt: 2022/11/7 19:16
describe: 白仪 https://www.zhihu.com/people/bai-yi-520/posts 知乎上定义的一些信号

byi 是 bai yi 的缩写
"""
import numpy as np
from typing import List
from czsc import CZSC
from collections import OrderedDict
from czsc.objects import BI, Direction, Mark
from czsc.utils import get_sub_elements, create_single_signal
from czsc.utils.sig import is_symmetry_zs
from czsc.signals.tas import update_macd_cache, update_boll_cache_V230228, update_ma_cache


def byi_symmetry_zs_V221107(c: CZSC, **kwargs):
    """对称中枢信号

    参数模板："{freq}_D{di}B_对称中枢"

    **信号逻辑：**

    从di笔往前数7/5/3笔，如果构成中枢，且所有笔的力度序列标
    准差小于均值的一定比例，则认为是对称中枢

    **信号列表：**

    - Signal('15分钟_D1B_对称中枢_否_向下_任意_0')
    - Signal('15分钟_D1B_对称中枢_是_向上_7笔_0')
    - Signal('15分钟_D1B_对称中枢_否_向上_任意_0')
    - Signal('15分钟_D1B_对称中枢_是_向下_3笔_0')
    - Signal('15分钟_D1B_对称中枢_是_向上_3笔_0')
    - Signal('15分钟_D1B_对称中枢_是_向下_5笔_0')
    - Signal('15分钟_D1B_对称中枢_是_向上_5笔_0')
    - Signal('15分钟_D1B_对称中枢_是_向下_7笔_0')

    :param c: CZSC对象
    :return: 信号识别结果
    """
    di = int(kwargs.get("di", 1))
    bis: List[BI] = get_sub_elements(c.bi_list, di=di, n=10)
    k1, k2, k3 = f"{c.freq.value}_D{di}B_对称中枢".split("_")
    v1 = '其他'
    if len(bis) < 7:
        return create_single_signal(k1=k1, k2=k2, k3=k3, v1=v1)

    for i in (7, 5, 3):
        v1 = is_symmetry_zs(bis[-i:])
        if v1:
            v3 = f"{i}笔"
            break
        else:
            v3 = "任意"

    v1 = "是" if v1 else "否"
    v2 = "向上" if bis[-1].direction == Direction.Down else "向下"
    return create_single_signal(k1=k1, k2=k2, k3=k3, v1=v1, v2=v2, v3=v3)


def byi_bi_end_V230106(c: CZSC, **kwargs) -> OrderedDict:
    """白仪分型停顿辅助笔结束判断

    参数模板："{freq}_D0停顿分型_BE辅助V230106"

    **信号逻辑：**

    分型停顿图解：https://pic1.zhimg.com/80/v2-3c5c3f264bffdf14c5ac6ae83bc5d5f0_720w.webp

    1. 白仪底分型停顿，认为是向下笔结束；反之，向上笔结束
    2. 底分型停顿：底分型后一根大阳线收盘在底分型的高点上方；反之，顶分型停顿

    **信号列表：**

    - Signal('15分钟_D0停顿分型_BE辅助V230106_看多_强_任意_0')
    - Signal('15分钟_D0停顿分型_BE辅助V230106_看空_强_任意_0')
    - Signal('15分钟_D0停顿分型_BE辅助V230106_看空_弱_任意_0')
    - Signal('15分钟_D0停顿分型_BE辅助V230106_看多_弱_任意_0')

    **Notes：**

    1. BE 是 Bi End 的缩写

    :param c: CZSC对象
    :return: 信号识别结果
    """
    k1, k2, k3 = f"{c.freq.value}_D0停顿分型_BE辅助V230106".split('_')
    v1 = "其他"

    if len(c.bi_list) < 3 or len(c.bars_ubi) > 7:
        return create_single_signal(k1=k1, k2=k2, k3=k3, v1=v1)

    last_bi = c.bi_list[-1]
    bars = get_sub_elements(c.bars_raw, di=1, n=3)
    last_fx = last_bi.fx_b
    bar1, bar2, bar3 = bars

    lc1 = last_bi.direction == Direction.Down and last_fx.mark == Mark.D and bar1.low == last_fx.low
    if lc1 and bar3.close > max([x.high for x in last_fx.raw_bars]):
        v1 = "看多"
        v2 = "强" if bar3.close > bar3.open and bar3.solid > max(bar3.upper, bar3.lower) else "弱"
        return create_single_signal(k1=k1, k2=k2, k3=k3, v1=v1, v2=v2)

    sc1 = last_bi.direction == Direction.Up and last_fx.mark == Mark.G and bar1.high == last_fx.high
    if sc1 and bar3.close < min([x.low for x in last_fx.raw_bars]):
        v1 = "看空"
        v2 = "强" if bar3.close < bar3.open and bar3.solid > max(bar3.upper, bar3.lower) else "弱"
        return create_single_signal(k1=k1, k2=k2, k3=k3, v1=v1, v2=v2)

    return create_single_signal(k1=k1, k2=k2, k3=k3, v1=v1)


def byi_bi_end_V230107(c: CZSC, **kwargs) -> OrderedDict:
    """白仪验证分型辅助判断笔结束

    参数模板："{freq}_D0验证分型_BE辅助V230107"

    **信号逻辑：**

    验证分型图解：https://pic1.zhimg.com/80/v2-80ac88269286707db98a5560107da4ec_720w.webp

    1. 白仪验证底分型，认为是向下笔结束；反之，向上笔结束

    **信号列表：**

    - Signal('15分钟_D0验证分型_BE辅助V230107_看空_强_任意_0')
    - Signal('15分钟_D0验证分型_BE辅助V230107_看空_弱_任意_0')
    - Signal('15分钟_D0验证分型_BE辅助V230107_看多_弱_任意_0')
    - Signal('15分钟_D0验证分型_BE辅助V230107_看多_强_任意_0')

    :param c: CZSC对象
    :return: 信号识别结果
    """
    k1, k2, k3 = f"{c.freq.value}_D0验证分型_BE辅助V230107".split('_')

    v1 = "其他"
    if len(c.bi_list) < 3 or len(c.bars_ubi) > 7:
        return create_single_signal(k1=k1, k2=k2, k3=k3, v1=v1)

    last_bi = c.bi_list[-1]
    fx1, fx2, fx3 = c.fx_list[-3], c.fx_list[-2], c.fx_list[-1]
    bar1 = c.bars_raw[-1]

    if not (last_bi.fx_b.dt == fx1.dt and fx1.mark == fx3.mark and bar1.dt == fx3.raw_bars[-1].dt):
        return create_single_signal(k1=k1, k2=k2, k3=k3, v1=v1)

    _high = max([x.close for x in fx1.raw_bars + fx2.raw_bars])
    _low = min([x.close for x in fx1.raw_bars + fx2.raw_bars])

    lc1 = bar1.solid > max(bar1.upper, bar1.lower) and bar1.close > bar1.open and bar1.close > _high
    if last_bi.direction == Direction.Down and fx1.mark == Mark.D and fx3.low > fx1.low:
        v1 = "看多"
        v2 = "强" if lc1 else "弱"
        return create_single_signal(k1=k1, k2=k2, k3=k3, v1=v1, v2=v2)

    sc1 = bar1.solid > max(bar1.upper, bar1.lower) and bar1.close < bar1.open and bar1.close < _low
    if last_bi.direction == Direction.Up and fx1.mark == Mark.G and fx3.high < fx1.high:
        v1 = "看空"
        v2 = "强" if sc1 else "弱"
        return create_single_signal(k1=k1, k2=k2, k3=k3, v1=v1, v2=v2)

    return create_single_signal(k1=k1, k2=k2, k3=k3, v1=v1)


def byi_second_bs_V230324(c: CZSC, **kwargs) -> OrderedDict:
    """白仪二类买卖点辅助V230324

    参数模板："{freq}_D{di}MACD{fastperiod}#{slowperiod}#{signalperiod}回抽零轴_BS2辅助V230324"

    参考资料：https://zhuanlan.zhihu.com/p/550719065
    由于文字描述的比较模糊，笔的算法也有差异，这里的实现和原文有一定出入

    参数模板："{freq}_D{di}MACD{fastperiod}#{slowperiod}#{signalperiod}回抽零轴_BS2辅助V230324"

    **信号逻辑：**

    1. 二买定义：
        a. 1,3,5笔的dif值都小于0，且1,3,5笔的dif值中最大值小于-2倍标准差，且8笔的dif值大于0，且9笔的dif值小于0.3倍标准差
        b. 第9笔向下

    2. 二卖定义：
        a. 1,3,5笔的dif值都大于0，且1,3,5笔的dif值中最小值大于2倍标准差，且8笔的dif值小于0，且9笔的dif值大于-0.3倍标准差
        b. 第9笔向上

    **信号列表：**

    - Signal('15分钟_D1MACD12#26#9回抽零轴_BS2辅助V230324_看空_任意_任意_0')
    - Signal('15分钟_D1MACD12#26#9回抽零轴_BS2辅助V230324_看多_任意_任意_0')

    :param c: CZSC对象
    :param di: 从倒数第几笔开始检查
    :return: 信号识别结果
    """
    di = int(kwargs.get("di", 1))
    cache_key = update_macd_cache(c, **kwargs)
    freq = c.freq.value
    k1, k2, k3 = f"{freq}_D{di}{cache_key}回抽零轴_BS2辅助V230324".split('_')
    v1 = "其他"
    if len(c.bi_list) < di + 10:
        return create_single_signal(k1=k1, k2=k2, k3=k3, v1=v1)

    b1, b2, b3, b4, b5, b6, b7, b8, b9 = get_sub_elements(c.bi_list, di=di, n=9)
    b1_dif = b1.fx_b.raw_bars[1].cache[cache_key]['dif']
    b3_dif = b3.fx_b.raw_bars[1].cache[cache_key]['dif']
    b5_dif = b5.fx_b.raw_bars[1].cache[cache_key]['dif']
    b8_dif = b8.fx_b.raw_bars[1].cache[cache_key]['dif']
    b9_dif = b9.fx_b.raw_bars[1].cache[cache_key]['dif']
    dif_std = np.std([x.cache[cache_key]['dif'] for x in b1.raw_bars])

    if b9.direction == Direction.Down and max(b1_dif, b3_dif, b5_dif) < 0 \
            and min(b1_dif, b3_dif, b5_dif) < -dif_std * 2 < dif_std * 1 < b8_dif \
            and abs(b9_dif) < dif_std * 0.3:
        v1 = "看多"

    if b9.direction == Direction.Up and min(b1_dif, b3_dif, b5_dif) > 0 \
            and max(b1_dif, b3_dif, b5_dif) > dif_std * 2 > -dif_std * 1 > b8_dif \
            and abs(b9_dif) < dif_std * 0.3:
        v1 = "看空"

    return create_single_signal(k1=k1, k2=k2, k3=k3, v1=v1)


def byi_fx_num_V230628(c: CZSC, **kwargs) -> OrderedDict:
    """白仪前面下跌或上涨一笔次级别笔结构数量满足条件；贡献者：谌意勇

    参数模板："{freq}_D{di}笔分型数大于{num}_BE辅助V230628"

    **信号逻辑：**

    对于采用分型停顿或者分型验证开开仓，前一笔内部次级别笔结构尽量带结构，
    此信号函数为当分型笔数量判断大于 num 为满足条件

    **信号列表：**

    - Signal('15分钟_D1笔分型数大于4_BE辅助V230628_向下_满足_任意_0')
    - Signal('15分钟_D1笔分型数大于4_BE辅助V230628_向上_满足_任意_0')
    
    :param c: CZSC对象
    :param di: 从倒数第几笔开始检查
    :param num: 前笔内部次级别笔数量
    :return: 信号识别结果
    """
    di = int(kwargs.get("di", 1))
    num = int(kwargs.get('num', 4))
    freq = c.freq.value
    k1, k2, k3 = f"{freq}_D{di}笔分型数大于{num}_BE辅助V230628".split('_')
    v1 = "其他"
    if len(c.bi_list) < di + 1 or len(c.bars_ubi) > 7:
        return create_single_signal(k1=k1, k2=k2, k3=k3, v1=v1)
    
    bi = c.bi_list[-di]
    v1 = bi.direction.value
    v2 = "满足" if len(bi.fxs) >= num else "其他"
    return create_single_signal(k1=k1, k2=k2, k3=k3, v1=v1, v2=v2)



# https://s0cqcxuy3p.feishu.cn/wiki/Pf1fw1woQi4iJikbKJmcYToznxb
import sys
sys.path.insert(0, r"D:\ZB\git_repo\waditu\czsc")

import czsc
import pandas as pd
import numpy as np
import plotly.express as px
from czsc import CzscTrader
from loguru import logger
from pathlib import Path
from typing import Union, AnyStr, Callable
from czsc.utils.stats import daily_performance, evaluate_pairs

czsc.welcome()


def get_ensemble_weight(trader: CzscTrader, method: Union[AnyStr, Callable] = 'mean'):
    """获取 CzscTrader 中所有 positions 按照 method 方法集成之后的权重

    :param trader: CzscTrader
        缠论交易者
    :param method: str or callable
        集成方法，可选值包括：'mean', 'max', 'min', 'vote'
        也可以传入自定义的函数，函数的输入为 dict，key 为 position.name，value 为 position.pos, 样例输入：
            {'多头策略A': 1, '多头策略B': 1, '空头策略A': -1}
    :param kwargs:
    :return: pd.DataFrame
        columns = ['dt', 'symbol', 'weight', 'price']    
    """
    logger.info(f"trader positions: {[p.name for p in trader.positions]}")

    dfp = pd.DataFrame()
    for p in trader.positions:
        p_pos = pd.DataFrame(p.holds)
        if dfp.empty:
            dfp = p_pos.copy()
        else:
            assert dfp['dt'].equals(p_pos['dt'])
            dfp = dfp.merge(p_pos[['dt', 'pos']], on='dt', how='left')
        dfp.rename(columns={'pos': p.name}, inplace=True)
    
    pos_cols = [c for c in dfp.columns if c not in ['dt', 'weight', 'price']]
    if callable(method):
        dfp['weight'] = dfp[pos_cols].apply(lambda x: method(x.to_dict()), axis=1)
    else:
        method = method.lower()
        if method == "mean":
            dfp['weight'] = dfp[pos_cols].mean(axis=1)
        elif method == "max":
            dfp['weight'] = dfp[pos_cols].max(axis=1)
        elif method == "min":
            dfp['weight'] = dfp[pos_cols].min(axis=1)
        elif method == "vote":
            dfp['weight'] = dfp[pos_cols].apply(lambda x: np.sign(np.sum(x)), axis=1)
        else:
            raise ValueError(f"method {method} not supported")
    
    dfp['symbol'] = trader.symbol
    logger.info(f"trader weight decribe: {dfp['weight'].describe().round(4).to_dict()}")
    return dfp[['dt', 'symbol', 'weight', 'price']].copy()


class WeightBacktest:
    """持仓权重回测"""

    def __init__(self, dfw, digits=2, **kwargs) -> None:
        """持仓权重回测
        
        :param dfw: pd.DataFrame, columns = ['dt', 'symbol', 'weight', 'price'], 持仓权重数据，
            其中 
                dt      为结束时间，
                symbol  为合约代码，
                weight  为持仓权重，
                price   为结束时间对应的交易价格，可以是当前K线的收盘价，或者下一根K线的开盘价，或者未来N根K线的TWAP、VWAP等
            
            数据样例如下：
            ===================  ========  ========  =======
            dt                   symbol      weight    price
            ===================  ========  ========  =======
            2019-01-02 09:01:00  DLi9001       0.5   961.695
            2019-01-02 09:02:00  DLi9001       0.25  960.72
            2019-01-02 09:03:00  DLi9001       0.25  962.669
            2019-01-02 09:04:00  DLi9001       0.25  960.72
            2019-01-02 09:05:00  DLi9001       0.25  961.695
            ===================  ========  ========  =======
            
        :param digits: int, 权重列保留小数位数
        :param kwargs:
            - fee_rate: float，单边交易成本，包括手续费与冲击成本, 默认为 0.0002
        """
        self.kwargs = kwargs
        self.dfw = dfw.copy()
        self.digits = digits
        self.fee_rate = kwargs.get('fee_rate', 0.0002)
        self.dfw['weight'] = self.dfw['weight'].round(digits)
        self.symbols = list(self.dfw['symbol'].unique().tolist())
        self.res_path = Path(kwargs.get('res_path', "weight_backtest"))
        self.res_path.mkdir(exist_ok=True, parents=True)    
        logger.add(self.res_path.joinpath("weight_backtest.log"), rotation="1 week")
        logger.info(f"持仓权重回测参数：digits={digits}, fee_rate={self.fee_rate}，res_path={self.res_path}，kwargs={kwargs}")

    def get_symbol_daily(self, symbol):
        """获取某个合约的每日收益率

        :param symbol: str，合约代码
        :return: pd.DataFrame，品种每日收益率，
            columns = ['date', 'symbol', 'edge', 'return', 'cost']
            其中
                date    为交易日，
                symbol  为合约代码，
                edge    为每日收益率，
                return  为每日收益率减去交易成本后的真实收益，
                cost    为交易成本
            
            数据样例如下：

                ==========  ========  ============  ============  =======
                date        symbol            edge        return     cost
                ==========  ========  ============  ============  =======
                2019-01-02  DLi9001    0.00230261    0.00195919   0.00085
                2019-01-03  DLi9001    0.00425589    0.00310589   0.00115
                2019-01-04  DLi9001   -0.0014209    -0.0024709    0.00105
                2019-01-07  DLi9001    0.000988305  -0.000111695  0.0011
                2019-01-08  DLi9001   -0.0004743    -0.0016243    0.00115
                ==========  ========  ============  ============  =======
        """
        dfs = self.dfw[self.dfw['symbol'] == symbol].copy()
        dfs['edge'] = dfs['weight'] * (dfs['price'].shift(-1) / dfs['price'] - 1)
        dfs['cost'] = abs(dfs['weight'].shift(1) - dfs['weight']) * self.fee_rate
        dfs['edge_post_fee'] = dfs['edge'] - dfs['cost']
        daily = dfs.groupby(dfs['dt'].dt.date).agg({'edge': 'sum', 'edge_post_fee': 'sum', 'cost': 'sum'}).reset_index()
        daily['symbol'] = symbol
        daily.rename(columns={'edge_post_fee': 'return', 'dt': 'date'}, inplace=True)
        daily = daily[['date', 'symbol', 'edge', 'return', 'cost']]
        return daily

    def get_symbol_pairs(self, symbol):
        """获取某个合约的开平交易记录"""
        dfs = self.dfw[self.dfw['symbol'] == symbol].copy()
        dfs['direction'] = 0
        dfs.loc[dfs['weight'] > 0, 'direction'] = 1
        dfs.loc[dfs['weight'] < 0, 'direction'] = -1
        dfs['volume'] = (dfs['weight'] * pow(10, self.digits)).astype(int) * dfs['direction']
        dfs['bar_id'] = list(range(1, len(dfs)+1))

        # 根据权重变化生成开平仓记录
        operates = []

        def __add_operate(dt, bar_id, volume, price, operate):
            for _ in range(abs(volume)):
                op = {'bar_id': bar_id, "dt": dt, "price": price, "operate": operate}
                operates.append(op)

        rows = dfs.to_dict(orient='records')

        # 处理第一个 row
        if rows[0]['volume'] > 0:
            __add_operate(rows[0]['dt'], rows[0]['bar_id'], rows[0]['volume'], rows[0]['price'], operate='开多')
        elif rows[0]['volume'] < 0:
            __add_operate(rows[0]['dt'], rows[0]['bar_id'], rows[0]['volume'], rows[0]['price'], operate='开空')

        # 处理后续 rows
        for row1, row2 in zip(rows[:-1], rows[1:]):
            if row1['volume'] >= 0 and row2['volume'] >= 0:
                # 多头仓位变化对应的操作
                if row2['volume'] > row1['volume']:
                    __add_operate(row2['dt'], row2['bar_id'], row2['volume'] - row1['volume'], row2['price'], operate='开多')
                elif row2['volume'] < row1['volume']:
                    __add_operate(row2['dt'], row2['bar_id'], row1['volume'] - row2['volume'], row2['price'], operate='平多')

            elif row1['volume'] <= 0 and row2['volume'] <= 0:
                # 空头仓位变化对应的操作
                if row2['volume'] > row1['volume']:
                    __add_operate(row2['dt'], row2['bar_id'], row1['volume'] - row2['volume'], row2['price'], operate='平空')
                elif row2['volume'] < row1['volume']:
                    __add_operate(row2['dt'], row2['bar_id'], row2['volume'] - row1['volume'], row2['price'], operate='开空')

            elif row1['volume'] >= 0 and row2['volume'] <= 0:
                # 多头转换成空头对应的操作
                __add_operate(row2['dt'], row2['bar_id'], row1['volume'], row2['price'], operate='平多')
                __add_operate(row2['dt'], row2['bar_id'], row2['volume'], row2['price'], operate='开空')
            
            elif row1['volume'] <= 0 and row2['volume'] >= 0:
                # 空头转换成多头对应的操作
                __add_operate(row2['dt'], row2['bar_id'], row1['volume'], row2['price'], operate='平空')
                __add_operate(row2['dt'], row2['bar_id'], row2['volume'], row2['price'], operate='开多')

        pairs, opens =[], []
        for op in operates:
            if op['operate'] in ['开多', '开空']:
                opens.append(op)
                continue

            assert op['operate'] in ['平多', '平空']
            open_op = opens.pop()
            if open_op['operate'] == '开多':
                p_ret = round((op['price'] - open_op['price']) / open_op['price'] * 10000, 2)
                p_dir = '多头'
            else:
                p_ret = round((open_op['price'] - op['price']) / open_op['price'] * 10000, 2)
                p_dir = '空头'
            pair = {"标的代码": symbol, "交易方向": p_dir,
                    "开仓时间": open_op['dt'], "平仓时间": op['dt'], 
                    "开仓价格": open_op['price'], "平仓价格": op['price'],
                    "持仓K线数": op['bar_id'] - open_op['bar_id'] + 1, 
                    "事件序列": f"{open_op['operate']} -> {op['operate']}",
                    "持仓天数": (op['dt'] - open_op['dt']).days,
                    "盈亏比例": p_ret}
            pairs.append(pair)
        df_pairs = pd.DataFrame(pairs)
        return df_pairs

    def backtest(self):
        """回测所有合约的收益率"""
        res = {}
        for symbol in self.symbols:
            daily = self.get_symbol_daily(symbol)
            pairs = self.get_symbol_pairs(symbol)
            res[symbol] = {"daily": daily, "pairs": pairs}

        pd.to_pickle(res, self.res_path.joinpath("res.pkl"))
        logger.info(f"回测结果已保存到 {self.res_path.joinpath('res.pkl')}")

        # 品种等权费后日收益率
        dret = pd.concat([v['daily'] for v in res.values()], ignore_index=True)
        dret = pd.pivot_table(dret, index='date', columns='symbol', values='return').fillna(0)
        dret['total'] = dret[list(res.keys())].mean(axis=1)
        logger.info(f"品种等权费后日收益率：{daily_performance(dret['total'])}")
        dret.to_excel(self.res_path.joinpath("daily_return.xlsx"), index=True)
        logger.info(f"品种等权费后日收益率已保存到 {self.res_path.joinpath('daily_return.xlsx')}")

        # 品种等权费后日收益率资金曲线绘制
        dret = dret.cumsum()
        fig = px.line(dret, y=dret.columns.to_list(), title="费后日收益率资金曲线")
        fig.for_each_trace(lambda trace: trace.update(visible=True if trace.name == 'total' else 'legendonly'))
        fig.write_html(self.res_path.joinpath("daily_return.html"))
        logger.info(f"费后日收益率资金曲线已保存到 {self.res_path.joinpath('daily_return.html')}")

        # 所有开平交易记录的表现
        dfp = pd.concat([v['pairs'] for v in res.values()], ignore_index=True)
        pairs_stats = evaluate_pairs(dfp)
        pairs_stats = {k: v for k, v in pairs_stats.items() if k in ['单笔收益', '持仓K线数', '交易胜率', '持仓天数']}
        logger.info(f"所有开平交易记录的表现：{pairs_stats}")
        czsc.save_json(pairs_stats, self.res_path.joinpath("pairs_stats.json"))
        logger.info(f"所有开平交易记录的表现已保存到 {self.res_path.joinpath('pairs_stats.json')}")

        return res


def test_ensemble():
    """从单个 trader 中获取持仓权重，然后回测"""
    trader = czsc.dill_load(r"D:\czsc_bi_datas\期货CTA投研\2019-01-01_2022-01-01_BE44E170\backtest_E497C9B5\traders\DLi9001.trader")

    def __ensemble_method(x):
        return (x['5分钟MACD多头T0'] + x['5分钟SMA#40多头T0']) / 2
    # dfw = get_ensemble_weight(trader, method='mean')
    dfw = get_ensemble_weight(trader, method=__ensemble_method)
    wb = WeightBacktest(dfw, digits=1, fee_rate=0.0002, res_path=r"C:\Users\zengb\Desktop\weight_example_vote")
    wb.backtest()


def test_ensemble_weight():
    """从持仓权重样例数据中回测"""
    dfw = pd.read_feather(r"C:\Users\zengb\Desktop\weight_example.feather")
    wb = WeightBacktest(dfw, digits=1, fee_rate=0.0002, res_path=r"C:\Users\zengb\Desktop\weight_example")
    res = wb.backtest()




# coding: utf-8
"""
使用 pyecharts 定制绘图模块

"""

from pyecharts import options as opts
from pyecharts.charts import HeatMap, Kline, Line, Bar, Scatter, Grid, Boxplot
from pyecharts.commons.utils import JsCode
from typing import List
import numpy as np
from czsc.objects import Operate
from .ta import SMA, MACD


def heat_map(data: List[dict],
             x_label: List[str] = None,
             y_label: List[str] = None,
             title: str = "热力图",
             width: str = "900px",
             height: str = "680px") -> HeatMap:
    """绘制热力图

    :param data: 用于绘制热力图的数据，示例如下
        [{'x': '0hour', 'y': '0day', 'heat': 11},
         {'x': '0hour', 'y': '1day', 'heat': 40},
         {'x': '0hour', 'y': '2day', 'heat': 38},
         {'x': '0hour', 'y': '3day', 'heat': 36},
         {'x': '0hour', 'y': '4day', 'heat': 11}]
    :param x_label: x轴标签
    :param y_label: y轴标签
    :param title: 图表标题
    :param width: 图表宽度
    :param height: 图表高度
    :return: 图表
    """

    value = [[s['x'], s['y'], s['heat']] for s in data]
    heat = [s['heat'] for s in data]

    if not x_label:
        x_label = sorted(list(set([s['x'] for s in data])))

    if not y_label:
        y_label = sorted(list(set([s['y'] for s in data])))

    vis_map_opts = opts.VisualMapOpts(pos_left="90%", pos_top="20%", min_=min(heat), max_=max(heat))
    title_opts = opts.TitleOpts(title=title)
    init_opts = opts.InitOpts(page_title=title, width=width, height=height)
    dz_inside = opts.DataZoomOpts(False, "inside", xaxis_index=[0], range_start=80, range_end=100)
    dz_slider = opts.DataZoomOpts(True, "slider", xaxis_index=[0], pos_top="96%", pos_bottom="0%",
                                  range_start=80, range_end=100)
    legend_opts = opts.LegendOpts(is_show=False)

    hm = HeatMap(init_opts=init_opts)
    hm.add_xaxis(x_label)
    hm.add_yaxis("heat", y_label, value, label_opts=opts.LabelOpts(is_show=True, position="inside"))
    hm.set_global_opts(title_opts=title_opts, visualmap_opts=vis_map_opts, legend_opts=legend_opts,
                       xaxis_opts=opts.AxisOpts(grid_index=0), datazoom_opts=[dz_inside, dz_slider])
    return hm


def kline_pro(kline: List[dict],
              fx: List[dict] = [],
              bi: List[dict] = [],
              xd: List[dict] = [],
              bs: List[dict] = [],
              title: str = "缠中说禅K线分析",
              t_seq: List[int] = [],
              width: str = "1400px",
              height: str = '580px') -> Grid:
    """绘制缠中说禅K线分析结果

    :param kline: K线
    :param fx: 分型识别结果
    :param bi: 笔识别结果
        {'dt': Timestamp('2020-11-26 00:00:00'),
          'fx_mark': 'd',
          'start_dt': Timestamp('2020-11-25 00:00:00'),
          'end_dt': Timestamp('2020-11-27 00:00:00'),
          'fx_high': 144.87,
          'fx_low': 138.0,
          'bi': 138.0}
    :param xd: 线段识别结果
    :param bs: 买卖点
    :param title: 图表标题
    :param t_seq: 均线系统
    :param width: 图表宽度
    :param height: 图表高度
    :return: 用Grid组合好的图表
    """
    # 配置项设置
    # ------------------------------------------------------------------------------------------------------------------
    bg_color = "#1f212d"  # 背景
    up_color = "#F9293E"
    down_color = "#00aa3b"

    init_opts = opts.InitOpts(bg_color=bg_color, width=width, height=height, animation_opts=opts.AnimationOpts(False))
    title_opts = opts.TitleOpts(title=title, pos_top="1%",
                                title_textstyle_opts=opts.TextStyleOpts(color=up_color, font_size=20),
                                subtitle_textstyle_opts=opts.TextStyleOpts(color=down_color, font_size=12))

    label_show_opts = opts.LabelOpts(is_show=True)
    label_not_show_opts = opts.LabelOpts(is_show=False)
    legend_not_show_opts = opts.LegendOpts(is_show=False)
    red_item_style = opts.ItemStyleOpts(color=up_color)
    green_item_style = opts.ItemStyleOpts(color=down_color)
    k_style_opts = opts.ItemStyleOpts(color=up_color, color0=down_color, border_color=up_color,
                                      border_color0=down_color, opacity=0.8)

    legend_opts = opts.LegendOpts(is_show=True, pos_top="1%", pos_left="30%", item_width=14, item_height=8,
                                  textstyle_opts=opts.TextStyleOpts(font_size=12, color="#0e99e2"))
    brush_opts = opts.BrushOpts(tool_box=["rect", "polygon", "keep", "clear"],
                                x_axis_index="all", brush_link="all",
                                out_of_brush={"colorAlpha": 0.1}, brush_type="lineX")

    axis_pointer_opts = opts.AxisPointerOpts(is_show=True, link=[{"xAxisIndex": "all"}])

    dz_inside = opts.DataZoomOpts(False, "inside", xaxis_index=[0, 1, 2], range_start=80, range_end=100)
    dz_slider = opts.DataZoomOpts(True, "slider", xaxis_index=[0, 1, 2], pos_top="96%",
                                  pos_bottom="0%", range_start=80, range_end=100)

    yaxis_opts = opts.AxisOpts(is_scale=True, min_="dataMin", max_="dataMax",
                               splitline_opts=opts.SplitLineOpts(is_show=False),
                               axislabel_opts=opts.LabelOpts(color="#c7c7c7", font_size=8, position="inside"))

    grid0_xaxis_opts = opts.AxisOpts(type_="category", grid_index=0, axislabel_opts=label_not_show_opts,
                                     split_number=20, min_="dataMin", max_="dataMax",
                                     is_scale=True, boundary_gap=False,
                                     splitline_opts=opts.SplitLineOpts(is_show=False),
                                     axisline_opts=opts.AxisLineOpts(is_on_zero=False))

    tool_tip_opts = opts.TooltipOpts(
        trigger="axis",
        axis_pointer_type="cross",
        background_color="rgba(245, 245, 245, 0.8)",
        border_width=1,
        border_color="#ccc",
        position=JsCode("""
                    function (pos, params, el, elRect, size) {
    					var obj = {top: 10};
    					obj[['left', 'right'][+(pos[0] < size.viewSize[0] / 2)]] = 30;
    					return obj;
    				}
                    """),
        textstyle_opts=opts.TextStyleOpts(color="#000"),
    )

    # 数据预处理
    # ------------------------------------------------------------------------------------------------------------------
    dts = [x['dt'] for x in kline]
    # k_data = [[x['open'], x['close'], x['low'], x['high']] for x in kline]
    k_data = [opts.CandleStickItem(name=i, value=[x['open'], x['close'], x['low'], x['high']])
              for i, x in enumerate(kline)]

    vol = []
    for i, row in enumerate(kline):
        item_style = red_item_style if row['close'] > row['open'] else green_item_style
        bar = opts.BarItem(name=i, value=row['vol'], itemstyle_opts=item_style, label_opts=label_not_show_opts)
        vol.append(bar)

    close = np.array([x['close'] for x in kline], dtype=np.double)
    diff, dea, macd = MACD(close)
    macd_bar = []
    for i, v in enumerate(macd.tolist()):
        item_style = red_item_style if v > 0 else green_item_style
        bar = opts.BarItem(name=i, value=round(v, 4), itemstyle_opts=item_style,
                           label_opts=label_not_show_opts)
        macd_bar.append(bar)

    diff = diff.round(4)
    dea = dea.round(4)

    # K 线主图
    # ------------------------------------------------------------------------------------------------------------------
    chart_k = Kline()
    chart_k.add_xaxis(xaxis_data=dts)
    chart_k.add_yaxis(series_name="Kline", y_axis=k_data, itemstyle_opts=k_style_opts)

    chart_k.set_global_opts(
        legend_opts=legend_opts,
        datazoom_opts=[dz_inside, dz_slider],
        yaxis_opts=yaxis_opts,
        tooltip_opts=tool_tip_opts,
        axispointer_opts=axis_pointer_opts,
        brush_opts=brush_opts,
        title_opts=title_opts,
        xaxis_opts=grid0_xaxis_opts
    )

    # 加入买卖点 - 多头操作 - 空头操作
    if bs:
        long_opens = {'i': [], 'val': []}
        long_exits = {'i': [], 'val': []}
        short_opens = {'i': [], 'val': []}
        short_exits = {'i': [], 'val': []}

        for op in bs:
            _dt = op['dt']
            _price = round(op['price'], 4)
            _info = f"{op['op_desc']} - 价格{_price}"

            if op['op'] in [Operate.LO]:
                long_opens['i'].append(_dt)
                long_opens['val'].append([_price, _info])

            if op['op'] in [Operate.LE]:
                long_exits['i'].append(_dt)
                long_exits['val'].append([_price, _info])

            if op['op'] in [Operate.SO]:
                short_opens['i'].append(_dt)
                short_opens['val'].append([_price, _info])

            if op['op'] in [Operate.SE]:
                short_exits['i'].append(_dt)
                short_exits['val'].append([_price, _info])

        chart_lo = (
            Scatter().add_xaxis(xaxis_data=long_opens['i']).add_yaxis(
                series_name="多头操作",
                y_axis=long_opens['val'],
                symbol_size=25,
                symbol='diamond',
                label_opts=opts.LabelOpts(is_show=False),
                itemstyle_opts=opts.ItemStyleOpts(color='#ff461f'),
                tooltip_opts=opts.TooltipOpts(
                    textstyle_opts=opts.TextStyleOpts(font_size=12),
                    formatter=JsCode("function (params) {return params.value[2];}")
                ),
            )
        )
        chart_le = (
            Scatter().add_xaxis(xaxis_data=long_exits['i']).add_yaxis(
                series_name="多头操作",
                y_axis=long_exits['val'],
                symbol_size=25,
                symbol='diamond',
                label_opts=opts.LabelOpts(is_show=False),
                itemstyle_opts=opts.ItemStyleOpts(color='#afdd22'),
                tooltip_opts=opts.TooltipOpts(
                    textstyle_opts=opts.TextStyleOpts(font_size=12),
                    formatter=JsCode("function (params) {return params.value[2];}")
                ),
            )
        )
        chart_so = (
            Scatter().add_xaxis(xaxis_data=short_opens['i']).add_yaxis(
                series_name="空头订单",
                y_axis=short_opens['val'],
                symbol_size=25,
                symbol='triangle',
                label_opts=opts.LabelOpts(is_show=False),
                itemstyle_opts=opts.ItemStyleOpts(color='#ff461f'),
                tooltip_opts=opts.TooltipOpts(
                    textstyle_opts=opts.TextStyleOpts(font_size=12),
                    formatter=JsCode("function (params) {return params.value[2];}")
                ),
            )
        )
        chart_se = (
            Scatter().add_xaxis(xaxis_data=short_exits['i']).add_yaxis(
                series_name="空头订单",
                y_axis=short_exits['val'],
                symbol_size=25,
                symbol='triangle',
                label_opts=opts.LabelOpts(is_show=False),
                itemstyle_opts=opts.ItemStyleOpts(color='#afdd22'),
                tooltip_opts=opts.TooltipOpts(
                    textstyle_opts=opts.TextStyleOpts(font_size=12),
                    formatter=JsCode("function (params) {return params.value[2];}")
                ),
            )
        )

        chart_k = chart_k.overlap(chart_lo)
        chart_k = chart_k.overlap(chart_le)
        chart_k = chart_k.overlap(chart_so)
        chart_k = chart_k.overlap(chart_se)

    # 均线图
    # ------------------------------------------------------------------------------------------------------------------
    chart_ma = Line()
    chart_ma.add_xaxis(xaxis_data=dts)
    if not t_seq:
        t_seq = [5, 13, 21]

    ma_keys = dict()
    for t in t_seq:
        ma_keys[f"MA{t}"] = SMA(close, timeperiod=t)

    for i, (name, ma) in enumerate(ma_keys.items()):
        chart_ma.add_yaxis(series_name=name, y_axis=ma, is_smooth=True,
                           symbol_size=0, label_opts=label_not_show_opts,
                           linestyle_opts=opts.LineStyleOpts(opacity=0.8, width=1))

    chart_ma.set_global_opts(xaxis_opts=grid0_xaxis_opts, legend_opts=legend_not_show_opts)
    chart_k = chart_k.overlap(chart_ma)

    # 缠论结果
    # ------------------------------------------------------------------------------------------------------------------
    if fx:
        fx_dts = [x['dt'] for x in fx]
        fx_val = [round(x['fx'], 2) for x in fx]
        chart_fx = Line()
        chart_fx.add_xaxis(fx_dts)
        chart_fx.add_yaxis(series_name="FX", y_axis=fx_val,
                           symbol="circle", symbol_size=6, label_opts=label_show_opts,
                           itemstyle_opts=opts.ItemStyleOpts(color="rgba(152, 147, 193, 1.0)", ))

        chart_fx.set_global_opts(xaxis_opts=grid0_xaxis_opts, legend_opts=legend_not_show_opts)
        chart_k = chart_k.overlap(chart_fx)

    if bi:
        bi_dts = [x['dt'] for x in bi]
        bi_val = [round(x['bi'], 2) for x in bi]
        chart_bi = Line()
        chart_bi.add_xaxis(bi_dts)
        chart_bi.add_yaxis(series_name="BI", y_axis=bi_val,
                           symbol="diamond", symbol_size=10, label_opts=label_show_opts,
                           itemstyle_opts=opts.ItemStyleOpts(color="rgba(184, 117, 225, 1.0)", ),
                           linestyle_opts=opts.LineStyleOpts(width=1.5))

        chart_bi.set_global_opts(xaxis_opts=grid0_xaxis_opts, legend_opts=legend_not_show_opts)
        chart_k = chart_k.overlap(chart_bi)

    if xd:
        xd_dts = [x['dt'] for x in xd]
        xd_val = [x['xd'] for x in xd]
        chart_xd = Line()
        chart_xd.add_xaxis(xd_dts)
        chart_xd.add_yaxis(series_name="XD", y_axis=xd_val,
                           symbol="triangle", symbol_size=10,
                           itemstyle_opts=opts.ItemStyleOpts(color="rgba(37, 141, 54, 1.0)", ))

        chart_xd.set_global_opts(xaxis_opts=grid0_xaxis_opts, legend_opts=legend_not_show_opts)
        chart_k = chart_k.overlap(chart_xd)

    # 成交量图
    # ------------------------------------------------------------------------------------------------------------------
    chart_vol = Bar()
    chart_vol.add_xaxis(dts)
    chart_vol.add_yaxis(series_name="Volume", y_axis=vol, bar_width='60%')
    chart_vol.set_global_opts(
        xaxis_opts=opts.AxisOpts(
            type_="category",
            grid_index=1,
            boundary_gap=False,
            axislabel_opts=opts.LabelOpts(is_show=True, font_size=8, color="#9b9da9"),
        ),
        yaxis_opts=yaxis_opts, legend_opts=legend_not_show_opts,
    )

    # MACD图
    # ------------------------------------------------------------------------------------------------------------------
    chart_macd = Bar()
    chart_macd.add_xaxis(dts)
    chart_macd.add_yaxis(series_name="MACD", y_axis=macd_bar, bar_width='60%')
    chart_macd.set_global_opts(
        xaxis_opts=opts.AxisOpts(
            type_="category",
            grid_index=2,
            axislabel_opts=opts.LabelOpts(is_show=False),
            splitline_opts=opts.SplitLineOpts(is_show=False),
        ),
        yaxis_opts=opts.AxisOpts(
            grid_index=2,
            split_number=4,
            axisline_opts=opts.AxisLineOpts(is_on_zero=False),
            axistick_opts=opts.AxisTickOpts(is_show=False),
            splitline_opts=opts.SplitLineOpts(is_show=False),
            axislabel_opts=opts.LabelOpts(is_show=True, color="#c7c7c7"),
        ),
        legend_opts=opts.LegendOpts(is_show=False),
    )

    line = Line()
    line.add_xaxis(dts)
    line.add_yaxis(series_name="DIFF", y_axis=diff, label_opts=label_not_show_opts, is_symbol_show=False,
                   linestyle_opts=opts.LineStyleOpts(opacity=0.8, width=1.0, color="#da6ee8"))
    line.add_yaxis(series_name="DEA", y_axis=dea, label_opts=label_not_show_opts, is_symbol_show=False,
                   linestyle_opts=opts.LineStyleOpts(opacity=0.8, width=1.0, color="#39afe6"))

    chart_macd = chart_macd.overlap(line)

    grid0_opts = opts.GridOpts(pos_left="0%", pos_right="1%", pos_top="12%", height="58%")
    grid1_opts = opts.GridOpts(pos_left="0%", pos_right="1%", pos_top="74%", height="8%")
    grid2_opts = opts.GridOpts(pos_left="0%", pos_right="1%", pos_top="86%", height="10%")

    grid_chart = Grid(init_opts)
    grid_chart.add(chart_k, grid_opts=grid0_opts)
    grid_chart.add(chart_vol, grid_opts=grid1_opts)
    grid_chart.add(chart_macd, grid_opts=grid2_opts)
    return grid_chart


def box_plot(data: dict,
             title: str = "箱线图",
             width: str = "900px",
             height: str = "680px") -> Boxplot:
    """

    :param data: 数据
        样例：
        data = {
            "expr 0": [960, 850, 830, 880],
            "expr 1": [960, 850, 830, 880],
        }
    :param title:
    :param width:
    :param height:
    :return:
    """
    x_data = []
    y_data = []
    for k, v in data.items():
        x_data.append(k)
        y_data.append(v)

    init_opts = opts.InitOpts(page_title=title, width=width, height=height)

    chart = Boxplot(init_opts=init_opts)
    chart.add_xaxis(xaxis_data=x_data)
    chart.add_yaxis(series_name="", y_axis=y_data)
    chart.set_global_opts(title_opts=opts.TitleOpts(pos_left="center", title=title),
                          tooltip_opts=opts.TooltipOpts(trigger="item", axis_pointer_type="shadow"),
                          xaxis_opts=opts.AxisOpts(
                              type_="category",
                              boundary_gap=True,
                              splitarea_opts=opts.SplitAreaOpts(is_show=False),
                              axislabel_opts=opts.LabelOpts(formatter="{value}"),
                              splitline_opts=opts.SplitLineOpts(is_show=False),
                          ),
                          yaxis_opts=opts.AxisOpts(
                              type_="value",
                              name="",
                              splitarea_opts=opts.SplitAreaOpts(
                                  is_show=True, areastyle_opts=opts.AreaStyleOpts(opacity=1)
                              )
                          ))
    return chart



# -*- coding: utf-8 -*-
"""
author: zengbin93
email: zeng_bin8888@163.com
create_dt: 2022/5/10 15:19
describe: 请描述文件用途
"""
import os
import pandas as pd
from loguru import logger
from czsc.objects import cal_break_even_point
import matplotlib.pyplot as plt
from czsc.data import TsDataCache, save_symbols_to_ebk
from czsc.sensors.utils import max_draw_down, turn_over_rate


plt.style.use('ggplot')
plt.rcParams['font.sans-serif'] = ['SimHei']
plt.rcParams['axes.unicode_minus'] = False


def stock_holds_performance(dc: TsDataCache, dfh, res_path):
    """计算A股日线持仓组合的表现

    :param dc: Tushare 数据缓存对象
    :param dfh: 持仓组合，样例如下，其中【证券代码】要求是tushare的格式，成分日期当天状态为持仓
            成分日期     证券代码    持仓权重     n1b
        0  2020-01-03  300620.SZ  0.008403  141.861099
        1  2020-01-03  300677.SZ  0.008403  767.124023
        2  2020-01-03  300708.SZ  0.008403   93.029297
        3  2020-01-03  002151.SZ  0.008403   -7.465500
        4  2020-01-03  002156.SZ  0.008403  350.101715
    :param res_path: 结果保存路径
    :return:
    """
    os.makedirs(res_path, exist_ok=True)
    date_fmt = '%Y%m%d'
    dfh['成分日期'] = pd.to_datetime(dfh['成分日期']).dt.strftime(date_fmt)
    sdt = pd.to_datetime(dfh['成分日期'].min()).strftime(date_fmt)
    edt = pd.to_datetime(dfh['成分日期'].max()).strftime(date_fmt)
    dfh.to_feather(os.path.join(res_path, f'holds_{sdt}_{edt}.feather'))
    latest = dfh[dfh['成分日期'] == dfh['成分日期'].max()]
    latest.to_excel(os.path.join(res_path, "组合最新持仓.xlsx"), index=False)
    save_symbols_to_ebk(latest['证券代码'].to_list(), os.path.join(res_path, "组合最新持仓.EBK"), source='ts')

    # 一进一出算1倍组合换手
    _, turn = turn_over_rate(dfh)
    mean_counts = round(dfh.groupby('成分日期')['证券代码'].count().mean(), 2)

    index_list = ['000905.SH', '000016.SH', '000300.SH']
    # 计算收益曲线
    dfa = pd.DataFrame({"成分日期": dc.get_dates_span(sdt, edt)})
    dfa['成分日期'] = pd.to_datetime(dfa['成分日期']).apply(lambda x: x.strftime(date_fmt))
    dfa = dfa.sort_values(by='成分日期')

    df_ = dfh.groupby('成分日期')['n1b'].mean().reset_index(drop=False)
    dfa = dfa.merge(df_[['成分日期', 'n1b']], on=['成分日期'], how='left')
    dfa.rename({'n1b': '组合收益'}, axis=1, inplace=True)

    for _index in index_list:
        dfi = dc.pro_bar(_index, sdt, edt, freq='D', asset="I", adj='qfq', raw_bar=False)
        dfi['成分日期'] = dfi['dt'].apply(lambda x: x.strftime(date_fmt))
        dfa = dfa.merge(dfi[['成分日期', 'n1b']], on=['成分日期'], how='left')
        dfa.rename({'n1b': _index}, axis=1, inplace=True)

    dfa = dfa.fillna(0)
    dfa.to_excel(os.path.join(res_path, f'收益对比_{sdt}_{edt}.xlsx'), index=False)
    mdd = max_draw_down(dfa['组合收益'])

    # 绘制收益曲线
    plt.close()
    fig = plt.figure(figsize=(13, 4 * len(index_list)))
    axes = fig.subplots(len(index_list), 1, sharex=True)
    for i, _index in enumerate(index_list):
        ax = axes[i]
        df_alpha = dfa.copy(deep=True).dropna(subset=['组合收益'])
        df_alpha['超额收益'] = df_alpha['组合收益'] - df_alpha[_index]
        df_alpha['组合收益'] = df_alpha['组合收益'].cumsum()
        df_alpha[_index] = df_alpha[_index].cumsum()
        df_alpha['超额收益'] = df_alpha['超额收益'].cumsum()
        df_alpha['成分日期'] = pd.to_datetime(df_alpha['成分日期']).apply(lambda x: x.date())

        ax.set_title(f"组合：日均持有{mean_counts}只股票，双边换手{round(turn, 2)}倍，最大回撤{int(mdd[2] * 10000)}BP")
        ax.plot(df_alpha['成分日期'], df_alpha['超额收益'], "r-", alpha=0.4)
        ax.plot(df_alpha['成分日期'], df_alpha['组合收益'], "b-", alpha=0.4)
        ax.plot(df_alpha['成分日期'], df_alpha[_index], "g-", alpha=0.4)
        ax.legend(['超额收益', '组合收益', f"基准：{_index}"], loc='upper left')
        ax.set_ylabel("净值（单位: BP）")
        plt.xticks(rotation=45)

    plt.tight_layout()
    file_png = os.path.join(res_path, f"alpha_plot.png")
    plt.savefig(file_png, bbox_inches='tight', dpi=100)
    plt.close()


class PairsPerformance:
    """交易对效果评估"""

    def __init__(self, df_pairs: pd.DataFrame, ):
        """

        :param df_pairs: 全部交易对，数据样例如下
                标的代码   交易方向 最大仓位  开仓时间         累计开仓                平仓时间  \
            0  000001.SH   多头     1 2020-02-06 09:45:00  2820.014893 2020-02-10 13:15:00
            1  000001.SH   多头     1 2020-03-20 14:15:00  2733.164062 2020-03-27 14:15:00
            2  000001.SH   多头     1 2020-03-30 13:30:00  2747.813965 2020-03-31 13:15:00
            3  000001.SH   多头     1 2020-04-01 10:45:00  2765.350098 2020-04-02 09:45:00
            4  000001.SH   多头     1 2020-04-02 14:15:00  2757.827881 2020-04-09 11:15:00
               累计平仓     累计换手  持仓K线数       事件序列           持仓天数    盈亏金额    交易盈亏  \
            0  2872.166992     2     40  开多@低吸 > 平多@60分钟顶背驰  4.145833  52.152100  0.0184
            1  2786.754883     2     80  开多@低吸 > 平多@60分钟顶背驰  7.000000  53.590820  0.0196
            2  2752.198975     2     15     开多@低吸 > 平多@持有资金  0.989583   4.385010  0.0015
            3  2721.335938     2     12     开多@低吸 > 平多@持有资金  0.958333 -44.014160 -0.0159
            4  2821.693115     2     58  开多@低吸 > 平多@60分钟顶背驰  6.875000  63.865234  0.0231
               盈亏比例
            0  0.0184
            1  0.0196
            2  0.0015
            3 -0.0159
            4  0.0231

        """
        df_pairs = df_pairs.copy(deep=True)
        # 将时间转换为年月日周
        time_convert = lambda x: (x.strftime("%Y年"), x.strftime("%Y年%m月"), x.strftime("%Y-%m-%d"),
                                  f"{x.year}年第{x.weekofyear}周" if x.weekofyear >= 10 else f"{x.year}年第0{x.weekofyear}周",
                                  )
        df_pairs[['开仓年', '开仓月', '开仓日', '开仓周']] = list(df_pairs['开仓时间'].apply(time_convert))
        df_pairs[['平仓年', '平仓月', '平仓日', '平仓周']] = list(df_pairs['平仓时间'].apply(time_convert))

        self.df_pairs = df_pairs
        # 指定哪些列可以用来进行聚合分析
        self.agg_columns = ['标的代码', '交易方向', '平仓年', '平仓月', '平仓周', '平仓日', '开仓年', '开仓月', '开仓日', '开仓周']

    @staticmethod
    def get_pairs_statistics(df_pairs: pd.DataFrame):
        """统计一组交易的基本信息

        :param df_pairs:
        :return:
        """
        if len(df_pairs) == 0:
            info = {
                "开始时间": None,
                "结束时间": None,
                "交易标的数量": 0,
                "总体交易次数": 0,
                "平均持仓天数": 0,

                "平均单笔收益": 0,
                "单笔收益标准差": 0,
                "最大单笔收益": 0,
                "最小单笔收益": 0,

                "交易胜率": 0,
                "累计盈亏比": 0,
                "单笔盈亏比": 0,
                "交易得分": 0,
                "赢面": 0,
                "每自然日收益": 0,
                "每根K线收益": 0,
                "盈亏平衡点": 0,
                "开仓日盈亏平衡点": 0,
            }
            return info

        win_pct = round(len(df_pairs[df_pairs['盈亏比例'] > 0]) / len(df_pairs), 4)
        df_gain = df_pairs[df_pairs['盈亏比例'] > 0]
        df_loss = df_pairs[df_pairs['盈亏比例'] <= 0]

        # 限制盈亏比最大有效值为 5
        single_gain_loss_rate = min(round(df_gain['盈亏比例'].mean() / (abs(df_loss['盈亏比例'].mean()) + 1e-8), 2), 5)
        total_gain_loss_rate = min(round(df_gain['盈亏比例'].sum() / (abs(df_loss['盈亏比例'].sum()) + 1e-8), 2), 5)

        info = {
            "开始时间": df_pairs['开仓时间'].min(),
            "结束时间": df_pairs['平仓时间'].max(),

            "交易标的数量": df_pairs['标的代码'].nunique(),
            "总体交易次数": len(df_pairs),
            "平均持仓天数": round(df_pairs['持仓天数'].mean(), 2),
            "平均持仓K线数": round(df_pairs['持仓K线数'].mean(), 2),

            "平均单笔收益": round(df_pairs['盈亏比例'].mean(), 4),
            "单笔收益标准差": round(df_pairs['盈亏比例'].std(), 4),
            "最大单笔收益": round(df_pairs['盈亏比例'].max(), 4),
            "最小单笔收益": round(df_pairs['盈亏比例'].min(), 4),

            "交易胜率": win_pct,
            "单笔盈亏比": single_gain_loss_rate,
            "累计盈亏比": total_gain_loss_rate,
            "交易得分": round(total_gain_loss_rate * win_pct, 4),
            "赢面": round(single_gain_loss_rate * win_pct - (1 - win_pct), 4),
            "盈亏平衡点": round(cal_break_even_point(df_pairs['盈亏比例'].to_list()), 4),
            "开仓日盈亏平衡点": round(df_pairs.groupby('开仓日')['盈亏比例'].apply(cal_break_even_point).mean(), 4),
        }

        info['每自然日收益'] = round(info['平均单笔收益'] / info['平均持仓天数'], 2)
        info['每根K线收益'] = round(info['平均单笔收益'] / info['平均持仓K线数'], 2)
        return info

    def agg_statistics(self, col: str):
        """按列聚合进行交易对评价"""
        df_pairs = self.df_pairs.copy()
        assert col in self.agg_columns, f"{col} 不是支持聚合的列，参考：{self.agg_columns}"

        results = []
        for name, dfg in df_pairs.groupby(col):
            if dfg.empty:
                continue

            res = {col: name}
            res.update(self.get_pairs_statistics(dfg))
            results.append(res)
        df = pd.DataFrame(results)
        return df

    @property
    def basic_info(self):
        """写入基础信息"""
        df_pairs = self.df_pairs.copy()
        return self.get_pairs_statistics(df_pairs)

    def agg_to_excel(self, file_xlsx):
        """遍历聚合列，保存结果到 Excel 文件中"""
        f = pd.ExcelWriter(file_xlsx)
        for col in ['标的代码', '交易方向', '平仓年', '平仓月', '平仓周', '平仓日']:
            df_ = self.agg_statistics(col)
            df_.to_excel(f, sheet_name=f"{col}聚合", index=False)
        f.close()
        logger.info(f"交易次数：{len(self.df_pairs)}; 聚合分析结果文件：{file_xlsx}")


def combine_holds_and_pairs(holds, pairs, results_path):
    """结合股票池和择时策略开平交易进行分析

    :param holds: 组合股票池数据，样例：
                 成分日期    证券代码       n1b      持仓权重
            0  2020-01-02  000001.SZ  183.758194  0.001232
            1  2020-01-02  000002.SZ -156.633896  0.001232
            2  2020-01-02  000063.SZ  310.296204  0.001232
            3  2020-01-02  000066.SZ -131.824997  0.001232
            4  2020-01-02  000069.SZ  -38.561699  0.001232

    :param pairs: 择时策略开平交易数据，数据格式如下
                标的代码 交易方向  最大仓位   开仓时间         累计开仓       平仓时间  \
            0  002698.SZ   多头     1 2015-01-12 13:30:00  24.02790 2015-01-13 09:45:00
            1  300031.SZ   多头     1 2015-01-12 10:30:00  53.87420 2015-01-13 09:45:00
            2  300046.SZ   多头     1 2015-01-12 10:15:00  41.35824 2015-01-13 09:45:00
            3  300076.SZ   多头     1 2015-01-12 10:30:00  57.84800 2015-01-13 09:45:00
            4  300099.SZ   多头     1 2015-01-12 10:15:00  62.57308 2015-01-13 09:45:00
                累计平仓  累计换手 持仓K线数 持仓天数  盈亏金额  交易盈亏  盈亏比例
            0  23.38150     2      7  0.843750 -0.64640 -0.0269 -0.0269
            1  52.71284     2     13  0.968750 -1.16136 -0.0215 -0.0215
            2  40.72068     2     14  0.979167 -0.63756 -0.0154 -0.0154
            3  55.45144     2     13  0.968750 -2.39656 -0.0414 -0.0414
            4  61.50528     2     14  0.979167 -1.06780 -0.0170 -0.0170

    :param results_path: 分析结果目录
    :return:
    """
    dfh = holds.copy()
    dfh['开仓日期'] = pd.to_datetime(dfh['成分日期'])
    dfh['标的代码'] = dfh['证券代码']

    dfp = pairs.copy()
    dfp['开仓日期'] = pd.to_datetime(dfp['开仓时间'].apply(lambda x: x.date()))

    # 合并，选择组合持仓权重大于 0 的交易对
    dfp_ = dfp.merge(dfh[['开仓日期', '标的代码', '持仓权重']], on=['开仓日期', '标的代码'], how='left')
    df_pairs = dfp_[dfp_['持仓权重'] > 0]

    # 按筛选出的交易对时间范围过滤原始交易对
    dfp_sub = dfp[(dfp['开仓时间'] >= df_pairs['开仓时间'].min()) & (dfp['开仓时间'] <= df_pairs['开仓时间'].max())]

    tp_old = PairsPerformance(dfp_sub)
    tp_new = PairsPerformance(df_pairs)
    print(f"原始交易：{tp_old.basic_info}，\n{tp_old.agg_statistics('平仓年')}\n")
    print(f"组合过滤：{tp_new.basic_info}，\n{tp_new.agg_statistics('平仓年')}")

    os.makedirs(results_path, exist_ok=True)
    tp_old.agg_to_excel(os.path.join(results_path, "原始交易评价.xlsx"))
    tp_new.agg_to_excel(os.path.join(results_path, "组合过滤评价.xlsx"))
    df_pairs.reset_index(drop=True, inplace=True)
    df_pairs.to_feather(os.path.join(results_path, "组合过滤交易.feather"))
    return tp_old, tp_new


def combine_dates_and_pairs(dates: list, pairs: pd.DataFrame, results_path):
    """结合大盘日期择时和择时策略开平交易进行分析

    :param dates: 大盘日期择时日期数据，数据样例 ['2020-01-02', ..., '2022-01-06']
    :param pairs: 择时策略开平交易数据，数据格式如下
                标的代码 交易方向  最大仓位   开仓时间         累计开仓       平仓时间  \
            0  002698.SZ   多头     1 2015-01-12 13:30:00  24.02790 2015-01-13 09:45:00
            1  300031.SZ   多头     1 2015-01-12 10:30:00  53.87420 2015-01-13 09:45:00
            2  300046.SZ   多头     1 2015-01-12 10:15:00  41.35824 2015-01-13 09:45:00
            3  300076.SZ   多头     1 2015-01-12 10:30:00  57.84800 2015-01-13 09:45:00
            4  300099.SZ   多头     1 2015-01-12 10:15:00  62.57308 2015-01-13 09:45:00
                累计平仓  累计换手 持仓K线数 持仓天数  盈亏金额  交易盈亏  盈亏比例
            0  23.38150     2      7  0.843750 -0.64640 -0.0269 -0.0269
            1  52.71284     2     13  0.968750 -1.16136 -0.0215 -0.0215
            2  40.72068     2     14  0.979167 -0.63756 -0.0154 -0.0154
            3  55.45144     2     13  0.968750 -2.39656 -0.0414 -0.0414
            4  61.50528     2     14  0.979167 -1.06780 -0.0170 -0.0170

    :param results_path: 分析结果目录
    :return:
    """
    dates = [pd.to_datetime(x) for x in dates]
    dfp = pairs.copy()
    dfp['开仓日期'] = pd.to_datetime(dfp['开仓时间'].apply(lambda x: x.date()))
    df_pairs = dfp[dfp['开仓日期'].isin(dates)]

    # 按筛选出的交易对时间范围过滤原始交易对
    dfp_sub = dfp[(dfp['开仓时间'] >= df_pairs['开仓时间'].min()) & (dfp['开仓时间'] <= df_pairs['开仓时间'].max())]

    tp_old = PairsPerformance(dfp_sub)
    tp_new = PairsPerformance(df_pairs)
    print(f"原始交易：{tp_old.basic_info}，\n{tp_old.agg_statistics('平仓年')}\n")
    print(f"组合过滤：{tp_new.basic_info}，\n{tp_new.agg_statistics('平仓年')}")

    os.makedirs(results_path, exist_ok=True)
    tp_old.agg_to_excel(os.path.join(results_path, "原始交易评价.xlsx"))
    tp_new.agg_to_excel(os.path.join(results_path, "组合过滤评价.xlsx"))
    return tp_old, tp_new



# -*- coding: utf-8 -*-
"""
author: zengbin93
email: zeng_bin8888@163.com
create_dt: 2023/3/13 22:45
describe: ZDY 信号函数集合
"""
import numpy as np
from loguru import logger
from collections import OrderedDict
from czsc.traders.base import CZSC, CzscTrader
from czsc.signals.tas import update_macd_cache
from czsc.utils import get_sub_elements, create_single_signal, sorted_freqs
from czsc.objects import Direction, Mark, Operate, ZS


def zdy_bi_end_V230406(c: CZSC, **kwargs) -> OrderedDict:
    """分型停顿判断K线结束

    参数模板："{freq}_D0停顿分型_BE辅助V230406"

    **信号逻辑：**

    1. 当分型形成后，等待后续出现某一根 K 线的收盘价站住第三根 K 线（此处的第三根 K 线是指形成分型的第三根 K 线）
    的极值，即可认为满足了停顿法的要求。
    2. 如果是两次停顿确认笔结束，要求内部至少还有一个相应的分型停顿。

    **信号列表：**

    - Signal('15分钟_D0停顿分型_BE辅助V230406_看空_任意_任意_0')
    - Signal('15分钟_D0停顿分型_BE辅助V230406_看多_任意_任意_0')
    - Signal('15分钟_D0停顿分型_BE辅助V230406_看多_内部底停顿_任意_0')
    - Signal('15分钟_D0停顿分型_BE辅助V230406_看空_内部顶停顿_任意_0')
    - Signal('15分钟_D0停顿分型_BE辅助V230406_看空_任意_顶分区间_0')
    - Signal('15分钟_D0停顿分型_BE辅助V230406_看多_任意_底分区间_0')
    - Signal('15分钟_D0停顿分型_BE辅助V230406_看空_内部顶停顿_顶分区间_0')
    - Signal('15分钟_D0停顿分型_BE辅助V230406_看多_内部底停顿_底分区间_0')

    **相关信号：**

    - :func:`czsc.signals.byi_bi_end_V230106`

    :param c: CZSC对象
    :return:
    """
    freq = c.freq.value
    k1, k2, k3 = f"{freq}_D0停顿分型_BE辅助V230406".split("_")
    v1 = '其他'
    if len(c.bi_list) < 3 or len(c.bars_ubi) > 6 or len(c.bars_ubi) < 4:
        return create_single_signal(k1=k1, k2=k2, k3=k3, v1=v1)

    last_bi = c.bi_list[-1]
    last_high = max([x.high for x in last_bi.fx_b.elements[-1].raw_bars])
    last_low = min([x.low for x in last_bi.fx_b.elements[-1].raw_bars])
    last_bar = c.bars_raw[-1]

    if last_bi.fx_b.elements[-1].dt >= last_bar.dt or last_bi.length < 7:
        return create_single_signal(k1=k1, k2=k2, k3=k3, v1=v1)

    last_bars = [y for x in c.bars_ubi[3:] for y in x.raw_bars]
    max_close = max([x.close for x in last_bars])
    min_close = min([x.close for x in last_bars])

    if last_bi.direction == Direction.Down and max_close > last_high:
        v1 = '看多'
    elif last_bi.direction == Direction.Up and min_close < last_low:
        v1 = '看空'

    if v1 == '其他':
        return create_single_signal(k1=k1, k2=k2, k3=k3, v1=v1)

    # 笔内部再次查找停顿分型
    v2 = "任意"
    if v1 == '看多' and len(last_bi.fxs) >= 4:
        for fx1, fx2 in zip(last_bi.fxs[:-1], last_bi.fxs[1:]):
            if (
                fx1.mark == Mark.D
                and fx2.mark == Mark.G
                and max([x.close for x in fx2.raw_bars]) > fx1.elements[-1].high
            ):
                v2 = '内部底停顿'

    if v1 == "看空" and len(last_bi.fxs) >= 4:
        for fx1, fx2 in zip(last_bi.fxs[:-1], last_bi.fxs[1:]):
            if (
                fx1.mark == Mark.G
                and fx2.mark == Mark.D
                and min([x.close for x in fx2.raw_bars]) < fx1.elements[-1].low
            ):
                v2 = '内部顶停顿'

    # 价格回到笔结束分型内部
    v3 = "任意"
    if v1 == '看多':
        assert last_bi.fx_b.mark == Mark.D
        if last_bar.close < last_bi.fx_b.high:
            v3 = '底分区间'
    if v1 == '看空':
        assert last_bi.fx_b.mark == Mark.G
        if last_bar.close > last_bi.fx_b.low:
            v3 = '顶分区间'

    return create_single_signal(k1=k1, k2=k2, k3=k3, v1=v1, v2=v2, v3=v3)


def zdy_bi_end_V230407(c: CZSC, **kwargs) -> OrderedDict:
    """分型停顿判断K线结束

    参数模板："{freq}_D0停顿分型_BE辅助V230407"

    **信号逻辑：**

    1. 当分型形成后，等待后续出现某一根 K 线的收盘价站住第三根 K 线（此处的第三根 K 线是指形成分型的第三根 K 线）
       的极值，且后续的所有K线收盘价都满足，即可认为满足了停顿法的要求。
    2. 如果是两次停顿确认笔结束，要求内部至少还有一个相应的分型停顿。

    **信号列表：**

    - Signal('15分钟_D0停顿分型_BE辅助V230407_看空_任意_任意_0')
    - Signal('15分钟_D0停顿分型_BE辅助V230407_看多_任意_任意_0')
    - Signal('15分钟_D0停顿分型_BE辅助V230407_看多_内部底停顿_任意_0')
    - Signal('15分钟_D0停顿分型_BE辅助V230407_看空_内部顶停顿_任意_0')

    :param c: CZSC对象
    :return:
    """
    freq = c.freq.value
    k1, k2, k3 = f"{freq}_D0停顿分型_BE辅助V230407".split("_")
    v1 = '其他'
    if len(c.bi_list) < 3 or len(c.bars_ubi) > 6 or len(c.bars_ubi) < 4:
        return create_single_signal(k1=k1, k2=k2, k3=k3, v1=v1)

    last_bi = c.bi_list[-1]
    last_high = max([x.high for x in last_bi.fx_b.elements[-1].raw_bars])
    last_low = min([x.low for x in last_bi.fx_b.elements[-1].raw_bars])
    last_bar = c.bars_raw[-1]

    if last_bi.fx_b.elements[-1].dt >= last_bar.dt or last_bi.length < 7:
        return create_single_signal(k1=k1, k2=k2, k3=k3, v1=v1)

    last_bars = [y for x in c.bars_ubi for y in x.raw_bars]
    last_bars = [x for x in last_bars if x.dt >= last_bi.fx_b.elements[-1].dt]

    if last_bi.direction == Direction.Down and last_bars[-1].close > last_high:
        _temp = [i for i, x in enumerate(last_bars) if x.close > last_high]
        if len(_temp) == 1 or (len(_temp) > 1 and _temp[-1] - _temp[0] == len(_temp) - 1):
            v1 = '看多'

    elif last_bi.direction == Direction.Up and last_bars[-1].close < last_low:
        _temp = [i for i, x in enumerate(last_bars) if x.close < last_low]
        if len(_temp) == 1 or (len(_temp) > 1 and _temp[-1] - _temp[0] == len(_temp) - 1):
            v1 = '看空'

    if v1 == '其他':
        return create_single_signal(k1=k1, k2=k2, k3=k3, v1=v1)

    # 笔内部再次查找停顿分型
    v2 = "任意"
    if v1 == '看多' and len(last_bi.fxs) >= 4:
        for fx1, fx2 in zip(last_bi.fxs[:-1], last_bi.fxs[1:]):
            if (
                fx1.mark == Mark.D
                and fx2.mark == Mark.G
                and max([x.close for x in fx2.raw_bars]) > fx1.elements[-1].high
            ):
                v2 = '内部底停顿'

    if v1 == "看空" and len(last_bi.fxs) >= 4:
        for fx1, fx2 in zip(last_bi.fxs[:-1], last_bi.fxs[1:]):
            if (
                fx1.mark == Mark.G
                and fx2.mark == Mark.D
                and min([x.close for x in fx2.raw_bars]) < fx1.elements[-1].low
            ):
                v2 = '内部顶停顿'

    return create_single_signal(k1=k1, k2=k2, k3=k3, v1=v1, v2=v2)


def zdy_vibrate_V230406(cat: CzscTrader, freq1, freq2, **kwargs) -> OrderedDict:
    """中枢震荡短差操作

    参数模板："中枢震荡_{freq1}#{freq2}_BS辅助V230406"

    **信号逻辑：**

    1、中枢笔数必须大于等于3笔。（排除掉盘背构成的中枢）
    2、中枢上沿做空，中枢下沿做多（不区分中枢方向）

    以开空仓为例（中枢上沿），当向上一笔出现次级别分型停顿后，定义次级别收盘价为 P，次级别顶分型高点为 H ，则需满足：

    1. H>=本级别中枢上沿价格
    2.（H-本级别中枢上沿价格）< 本级别中枢高度
    3.（H-P）*3  < 中枢高度
    4. 本级别MACD黄白线死叉确立（已经走出并列向下走势）

    稍微解释一下几个条件的内在逻辑：

    1. H>=本级别中枢上沿价格 ： 确保是在中枢的上沿
    2.（H-本级别中枢上沿价格）< 本级别中枢高度 ： 确保中枢上沿不要太远的距离，太远了可能下跌就形成三买上去了。这种点不要。
    3. （H-P）*3  < 中枢高度 ： H-P，其实就是一个次级别分型停顿的高度了，这里就是为了排除掉奔走型中枢，区间特别小的中枢，没必要做。

    **信号列表：**

    - Signal('中枢震荡_5分钟#60分钟_BS辅助V230406_看多_任意_任意_0')
    - Signal('中枢震荡_5分钟#60分钟_BS辅助V230406_看空_任意_任意_0')

    :param cat: 交易员对象
    :param freq1: 次级别
    :param freq2: 本级别
    :param kwargs:
    :return:
    """
    k1, k2, k3 = f"中枢震荡_{freq1}#{freq2}_BS辅助V230406".split("_")
    v1 = '其他'
    assert sorted_freqs.index(freq1) < sorted_freqs.index(freq2), "freq1 必须小于 freq2"
    c1: CZSC = cat.kas[freq1]
    c2: CZSC = cat.kas[freq2]
    cache_key = update_macd_cache(c2, fastperiod=12, slowperiod=26, signalperiod=9)

    if len(c2.bi_list) < 5:
        return create_single_signal(k1=k1, k2=k2, k3=k3, v1=v1)

    b1, b2, b3, b4 = c2.bi_list[-4:]
    zg = min(b1.high, b2.high, b3.high)
    zd = max(b1.low, b2.low, b3.low)

    if zd > zg:  # 本级别中枢不成立，不做
        return create_single_signal(k1=k1, k2=k2, k3=k3, v1=v1)

    # 判断次级别分型是否停顿
    c1_lbi = c1.bi_list[-1]
    c1_bar = c1.bars_raw[-1]

    if c1_bar.dt == c1_lbi.fx_b.raw_bars[-1].dt or len(c1.bars_ubi) > 6:
        # 次级别不具备分型停顿的条件，不做
        return create_single_signal(k1=k1, k2=k2, k3=k3, v1=v1)

    if (
        c1_lbi.direction == Direction.Down
        and c1_bar.close > c1_bar.open
        and c1_bar.close > c1_lbi.fx_b.raw_bars[-1].high
    ):
        temp = '底分停顿'

    elif (
        c1_lbi.direction == Direction.Up and c1_bar.close < c1_bar.open and c1_bar.close < c1_lbi.fx_b.raw_bars[-1].low
    ):
        temp = '顶分停顿'

    else:
        # 次级别分型没有停顿，不做
        return create_single_signal(k1=k1, k2=k2, k3=k3, v1=v1)

    # 判断中枢震荡买卖点
    c2_bar = c2.bars_raw[-1]
    if temp == '顶分停顿' and c2_bar.cache[cache_key]['macd'] < 0:
        P = c1_bar.close
        H = c1_lbi.high
        if H >= zg and (H - zg) < (zg - zd) and (H - P) * 3 < (zg - zd):
            v1 = '看空'

    if temp == '底分停顿' and c2_bar.cache[cache_key]['macd'] > 0:
        P = c1_bar.close
        L = c1_lbi.low
        if L <= zd and (zd - L) < (zg - zd) and (P - L) * 3 < (zg - zd):
            v1 = '看多'

    return create_single_signal(k1=k1, k2=k2, k3=k3, v1=v1)


def zdy_stop_loss_V230406(cat: CzscTrader, **kwargs) -> OrderedDict:
    """笔操作止损逻辑

    参数模板："{freq1}_{pos_name}F{first_stop}_止损V230406"

    **信号逻辑：**

    多头止损逻辑如下，反之为空头止损逻辑：

    1. 进场后设定止损为固定BP止损。
    2. 跌破开仓前最后一个分型的低点，止损。
    3. 任何一笔在持仓状态下的上升笔结束后，若没有判断止盈走人，则将止损设定在该笔起始点的位置（或适当放低一些固定BP）

    **信号列表：**

    - Signal('日线_5日线多头F300_止损V230406_多头止损_任意_任意_0')
    - Signal('日线_5日线多头F300_止损V230406_空头止损_任意_任意_0')

    :param cat: CzscTrader对象
    :return:
    """
    pos_name = kwargs["pos_name"]
    freq1 = kwargs["freq1"]  # 笔的观察周期
    first_stop = int(kwargs.get("first_stop", 300))  # 进场点止损设置
    k1, k2, k3 = f"{freq1}_{pos_name}F{first_stop}_止损V230406".split("_")
    v1, v2 = '其他', '其他'
    if hasattr(cat, 'positions') is False:
        return create_single_signal(k1=k1, k2=k2, k3=k3, v1=v1)

    pos = [x for x in cat.positions if x.name == pos_name][0]
    if len(pos.operates) == 0 or pos.operates[-1]['op'] in [Operate.SE, Operate.LE]:
        return create_single_signal(k1=k1, k2=k2, k3=k3, v1=v1)

    c = cat.kas[freq1]
    op = pos.operates[-1]
    d3bi = c.bi_list[-3]
    bis = [x for x in c.bi_list if x.fx_b.dt >= op['dt']]
    last_bar = c.bars_raw[-1]

    # 多头止损逻辑
    if op['op'] == Operate.LO:
        open_base_fx = [x for x in c.fx_list if x.mark == Mark.D and x.dt < op['dt']][-1]
        assert open_base_fx.mark == Mark.D, "开仓前最后一个分型必须是底分型"
        if last_bar.close < open_base_fx.low:
            v1 = '多头止损'
            v2 = '跌破分型低点'

        if (last_bar.close / op['price'] - 1) * 10000 <= -first_stop:
            v1 = '多头止损'
            v2 = '进场点止损'

        if (
            len(bis) > 0
            and bis[-1].direction == Direction.Up
            and bis[-1].high > d3bi.high
            and last_bar.close < op['price']
        ):
            assert bis[-1].direction == Direction.Up
            v1 = '多头止损'
            v2 = '跌破成本价'

        if len(bis) > 1 and bis[-1].direction == Direction.Up and last_bar.close < bis[-2].fx_b.low:
            v1 = '多头止损'
            v2 = '跌破上个向下笔底'

    # 空头止损逻辑
    if op['op'] == Operate.SO:
        open_base_fx = [x for x in c.fx_list if x.mark == Mark.G and x.dt < op['dt']][-1]
        assert open_base_fx.mark == Mark.G, "开仓前最后一个分型必须是底分型"
        if last_bar.close > open_base_fx.high:
            v1 = '空头止损'
            v2 = '升破分型高点'

        if (1 - last_bar.close / op['price']) * 10000 <= -first_stop:
            v1 = '空头止损'
            v2 = '进场点止损'

        if (
            len(bis) > 0
            and bis[-1].direction == Direction.Down
            and bis[-1].low < d3bi.low
            and last_bar.close > op['price']
        ):
            assert d3bi.direction == Direction.Down
            v1 = '空头止损'
            v2 = '升破成本价'

        if len(bis) > 1 and bis[-1].direction == Direction.Down and last_bar.close > bis[-2].fx_b.high:
            v1 = '空头止损'
            v2 = '升破上个向上笔顶'

    return create_single_signal(k1=k1, k2=k2, k3=k3, v1=v1, v2=v2)


def zdy_take_profit_V230406(cat: CzscTrader, **kwargs) -> OrderedDict:
    """笔操作止盈逻辑

    参数模板："{freq1}_{pos_name}_止盈V230406"

    **信号逻辑：**

    多头止盈逻辑如下，反之为空头止盈逻辑：
    
    1. 任何一笔在持仓状态下的上升笔结束，若升破前一下跌笔高点，继续持仓，如没有升破前一下跌笔高点，止盈走人。需要用两次停顿分型判断。

    **信号列表：**

    - Signal('日线_5日线多头_止盈V230406_多头止盈_任意_任意_0')
    - Signal('日线_5日线多头_止盈V230406_空头止盈_任意_任意_0')

    :param cat: CzscTrader对象
    :return:
    """
    pos_name = kwargs["pos_name"]
    freq1 = kwargs["freq1"]  # 笔的观察周期
    k1, k2, k3 = f"{freq1}_{pos_name}_止盈V230406".split("_")
    v1, v2 = '其他', '其他'
    if not hasattr(cat, "positions"):
        return create_single_signal(k1=k1, k2=k2, k3=k3, v1=v1)

    pos = [x for x in cat.positions if x.name == pos_name][0]
    if len(pos.operates) == 0 or pos.operates[-1]['op'] in [Operate.SE, Operate.LE]:
        return create_single_signal(k1=k1, k2=k2, k3=k3, v1=v1)

    op = pos.operates[-1]
    c = cat.kas[freq1]
    bis = [x for x in c.bi_list if x.fx_b.dt >= op['dt']]

    # 多头止盈逻辑
    if op['op'] == Operate.LO:
        if len(bis) > 1 and bis[-1].direction == Direction.Up and bis[-1].high < bis[-2].high:
            v1 = '多头止盈'
            v2 = '向上笔不创新高'

    # 空头止盈逻辑
    if op['op'] == Operate.SO:
        if len(bis) > 1 and bis[-1].direction == Direction.Down and bis[-1].low > bis[-2].low:
            v1 = '空头止盈'
            v2 = '向下笔不创新低'

    return create_single_signal(k1=k1, k2=k2, k3=k3, v1=v1, v2=v2)


def zdy_take_profit_V230407(cat: CzscTrader, **kwargs) -> OrderedDict:
    """笔操作止盈逻辑

    参数模板："{freq1}_{pos_name}_止盈V230407"

    **信号逻辑：**

    多头止盈逻辑如下，反之为空头止盈逻辑：
    1. 根据K线个数判断力度，提前止盈。（三买之后的上升笔中，如果笔数已经大于前一下跌笔的K线个数的1.5倍，并且还没有新高，直接止盈走人。

    **信号列表：**

    - Signal('日线_5日线多头_止盈V230407_多头止盈_任意_任意_0')
    - Signal('日线_5日线多头_止盈V230407_空头止盈_任意_任意_0')

    :param cat: CzscTrader对象
    :return: 信号字典
    """
    pos_name = kwargs["pos_name"]
    freq1 = kwargs["freq1"]  # 笔的观察周期
    k1, k2, k3 = f"{freq1}_{pos_name}_止盈V230407".split("_")
    v1, v2 = '其他', '其他'
    if not hasattr(cat, "positions"):
        return create_single_signal(k1=k1, k2=k2, k3=k3, v1=v1)

    pos = [x for x in cat.positions if x.name == pos_name][0]
    if len(pos.operates) == 0 or pos.operates[-1]['op'] in [Operate.SE, Operate.LE]:
        return create_single_signal(k1=k1, k2=k2, k3=k3, v1=v1)

    op = pos.operates[-1]
    c = cat.kas[freq1]
    bis = [x for x in c.bi_list if x.fx_b.dt >= op['dt']]
    d2bi = c.bi_list[-2]

    if len(bis) <= 0 or bis[-1].length < 1.5 * d2bi.length:
        return create_single_signal(k1=k1, k2=k2, k3=k3, v1=v1)

    # 多头止盈逻辑
    if op['op'] == Operate.LO and bis[-1].direction == Direction.Up and bis[-1].high < d2bi.high:
        assert d2bi.direction == Direction.Down
        v1 = '多头止盈'
        v2 = '向上笔不创新高'

    # 空头止盈逻辑
    if op['op'] == Operate.SO and bis[-1].direction == Direction.Down and bis[-1].low > d2bi.low:
        assert d2bi.direction == Direction.Up
        v1 = '空头止盈'
        v2 = '向下笔不创新低'

    return create_single_signal(k1=k1, k2=k2, k3=k3, v1=v1, v2=v2)


def zdy_zs_V230423(c: CZSC, **kwargs):
    """约束中枢的形态和高度

    参数模板："{freq}_D{di}中枢形态_BS辅助V230423"

    **信号逻辑：**

    以上涨中枢为例，反之为下跌中枢：

    1. 进入笔和离开笔分别是高低点，笔的数量大于等于5
    2. 中枢高度 >= 进入中枢那一笔高度 / 3

    **信号列表：**

    - Signal('15分钟_D1中枢形态_BS辅助V230423_上涨_5笔_任意_0')
    - Signal('15分钟_D1中枢形态_BS辅助V230423_上涨_7笔_任意_0')
    - Signal('15分钟_D1中枢形态_BS辅助V230423_下跌_5笔_任意_0')
    - Signal('15分钟_D1中枢形态_BS辅助V230423_下跌_7笔_任意_0')
    - Signal('15分钟_D1中枢形态_BS辅助V230423_下跌_9笔_任意_0')
    - Signal('15分钟_D1中枢形态_BS辅助V230423_上涨_9笔_任意_0')

    :param c: 基础周期的 CZSC 对象
    :param kwargs: 其他参数

        - di: 倒数第 di 根 K 线
        
    :return: 信号字典
    """
    di = int(kwargs.get('di', 1))
    freq = c.freq.value
    k1, k2, k3 = f"{freq}_D{di}中枢形态_BS辅助V230423".split('_')
    v1 = '其他'
    if len(c.bi_list) < 7 or len(c.bars_ubi) > 7:
        return create_single_signal(k1=k1, k2=k2, k3=k3, v1=v1)

    for n in (9, 7, 5):
        bis = get_sub_elements(c.bi_list, di=di, n=n)
        if len(bis) != n:
            continue

        bi1 = bis[0]
        # 假定离开中枢的都是一笔
        zs = ZS(bis[1:-1])
        if not (zs.is_valid and zs.zg - zs.zd > (bi1.high - bi1.low) / 3):
            continue

        min_low = min(x.low for x in bis)
        max_high = max(x.high for x in bis)
        if bi1.direction == Direction.Up and bi1.low == min_low and bis[-1].high == max_high:
            return create_single_signal(k1=k1, k2=k2, k3=k3, v1="上涨", v2=f"{n}笔")

        if bi1.direction == Direction.Down and bi1.high == max_high and bis[-1].low == min_low:
            return create_single_signal(k1=k1, k2=k2, k3=k3, v1="下跌", v2=f"{n}笔")

    return create_single_signal(k1=k1, k2=k2, k3=k3, v1=v1)


def zdy_macd_bc_V230422(c: CZSC, **kwargs):
    """MACD面积背驰

    参数模板："{freq}_D{di}T{th}MACD面积背驰_BS辅助V230422"

    **信号逻辑：**

    以上涨背驰为例，反之为下跌背驰：

    1. 背驰段的相应macd面积之和 <= 进入中枢段的相应面积之和 * th / 100
    2. 中枢把黄白线拉到0轴附近，
    3. 离开中枢的一笔，黄白线大于0且不新高

    **信号列表：**

    - Signal('15分钟_D1T50MACD面积背驰_BS辅助V230422_上涨_9笔_任意_0')
    - Signal('15分钟_D1T50MACD面积背驰_BS辅助V230422_上涨_7笔_任意_0')
    - Signal('15分钟_D1T50MACD面积背驰_BS辅助V230422_下跌_5笔_任意_0')
    - Signal('15分钟_D1T50MACD面积背驰_BS辅助V230422_上涨_5笔_任意_0')
    - Signal('15分钟_D1T50MACD面积背驰_BS辅助V230422_下跌_7笔_任意_0')
    - Signal('15分钟_D1T50MACD面积背驰_BS辅助V230422_下跌_9笔_任意_0')

    :param c: 基础周期的 CZSC 对象
    :param kwargs: 其他参数

        - di: 倒数第 di 根 K 线
        - th: 背驰段的相应macd面积之和 <= 进入中枢段的相应面积之和 * th / 100

    :return: 信号字典
    """
    cache_key = update_macd_cache(c, fastperiod=26, slowperiod=12, signalperiod=9)
    di = int(kwargs.get('di', 1))
    th = int(kwargs.get('th', 50))
    freq = c.freq.value
    k1, k2, k3 = f"{freq}_D{di}T{th}MACD面积背驰_BS辅助V230422".split('_')
    v1 = '其他'
    if len(c.bi_list) < 7 or len(c.bars_ubi) > 7:
        return create_single_signal(k1=k1, k2=k2, k3=k3, v1=v1)

    for n in (9, 7, 5):
        bis = get_sub_elements(c.bi_list, di=di, n=n)
        if len(bis) != n:
            continue

        # 假定离开中枢的都是一笔
        zs = ZS(bis[1:-1])
        if not zs.is_valid:  # 如果中枢不成立，往下进行
            continue

        bi1, bi2 = bis[0], bis[-1]
        bi1_macd = [x.cache[cache_key]['macd'] for x in bi1.raw_bars[1:-1]]
        bi2_macd = [x.cache[cache_key]['macd'] for x in bi2.raw_bars[1:-1]]
        bi1_dif = bi1.raw_bars[-2].cache[cache_key]['dif']
        bi2_dif = bi2.raw_bars[-2].cache[cache_key]['dif']

        zs_fxb_raw = [y for x in zs.bis for y in x.fx_b.raw_bars]

        if bi1.direction == Direction.Up:
            bi1_area = sum([x for x in bi1_macd if x > 0])
            bi2_area = sum([x for x in bi2_macd if x > 0])
            dif_zero = min([x.cache[cache_key]['dif'] for x in zs_fxb_raw])
        else:
            bi1_area = sum([x for x in bi1_macd if x < 0])
            bi2_area = sum([x for x in bi2_macd if x < 0])
            dif_zero = max([x.cache[cache_key]['dif'] for x in zs_fxb_raw])

        if bi2_area > bi1_area * th / 100:  # 如果面积背驰不成立，往下进行
            continue

        min_low = min(x.low for x in bis)
        max_high = max(x.high for x in bis)
        if (
            bi1.direction == Direction.Up
            and bi1.low == min_low
            and bi2.high == max_high
            and dif_zero < 0
            and bi1_dif > bi2_dif > 0
        ):
            return create_single_signal(k1=k1, k2=k2, k3=k3, v1="上涨", v2=f"{n}笔")

        if (
            bi1.direction == Direction.Down
            and bi1.high == max_high
            and bi2.low == min_low
            and dif_zero > 0
            and bi1_dif < bi2_dif < 0
        ):
            return create_single_signal(k1=k1, k2=k2, k3=k3, v1="下跌", v2=f"{n}笔")

    return create_single_signal(k1=k1, k2=k2, k3=k3, v1=v1)


def zdy_macd_bs1_V230422(c: CZSC, **kwargs):
    """MACD辅助判断第一类买卖点

    参数模板："{freq}_D{di}T{th}MACD_BS1辅助V230422"

    **信号逻辑：**

    以上涨背驰为例，反之为下跌背驰：

    1. 背驰段的相应macd面积之和 <= 进入中枢段的相应面积之和 * th / 100
    2. 离开中枢的一笔，起点的黄白线在零轴附近，终点的黄白线在中枢黄白线的上方

    **信号列表：**

    - Signal('5分钟_D1T50MACD_BS1辅助V230422_看空_上涨5笔_任意_0')
    - Signal('5分钟_D1T50MACD_BS1辅助V230422_看多_下跌7笔_任意_0')
    - Signal('5分钟_D1T50MACD_BS1辅助V230422_看多_下跌5笔_任意_0')
    - Signal('5分钟_D1T50MACD_BS1辅助V230422_看空_上涨9笔_任意_0')
    - Signal('5分钟_D1T50MACD_BS1辅助V230422_看空_上涨7笔_任意_0')
    - Signal('5分钟_D1T50MACD_BS1辅助V230422_看多_下跌9笔_任意_0')

    :param c: 基础周期的 CZSC 对象
    :param kwargs: 其他参数

        - di: 倒数第 di 根 K 线
        - th: 背驰段的相应macd面积之和 <= 进入中枢段的相应面积之和 * th / 100

    :return: 信号字典
    """
    cache_key = update_macd_cache(c, fastperiod=26, slowperiod=12, signalperiod=9)
    di = int(kwargs.get('di', 1))
    th = int(kwargs.get('th', 50))
    freq = c.freq.value
    k1, k2, k3 = f"{freq}_D{di}T{th}MACD_BS1辅助V230422".split('_')
    v1 = '其他'
    if len(c.bi_list) < 7 or len(c.bars_ubi) > 9:
        return create_single_signal(k1=k1, k2=k2, k3=k3, v1=v1)

    for n in (13, 11, 9, 7, 5):
        bis = get_sub_elements(c.bi_list, di=di, n=n)
        if len(bis) != n:
            continue

        # 假定离开中枢的都是一笔
        zs = ZS(bis[1:-1])
        if not zs.is_valid:  # 如果中枢不成立，往下进行
            continue

        bi1, bi2 = bis[0], bis[-1]
        bi1_area = sum([abs(x.cache[cache_key]['macd']) for x in bi1.raw_bars[1:-1]])
        bi2_area = sum([abs(x.cache[cache_key]['macd']) for x in bi2.raw_bars[1:-1]])
        bi1_dif = bi1.raw_bars[-2].cache[cache_key]['dif']
        bi2_dif = bi2.raw_bars[-2].cache[cache_key]['dif']
        bi2_start_dif = bi2.raw_bars[1].cache[cache_key]['dif']

        if bi1.direction == Direction.Up:
            # 计算 zs 中向上笔的 dif 最大值
            up_bis = [x for x in zs.bis if x.direction == Direction.Up]
            zs_dif = max([y.cache[cache_key]['dif'] for x in up_bis for y in x.fx_b.raw_bars])
        else:
            # 计算 zs 中向下笔的 dif 最小值
            down_bis = [x for x in zs.bis if x.direction == Direction.Down]
            zs_dif = min([y.cache[cache_key]['dif'] for x in down_bis for y in x.fx_b.raw_bars])

        if bi2_area > bi1_area * th / 100:  # 如果面积背驰不成立，往下进行
            continue

        min_low = min(x.low for x in bis)
        max_high = max(x.high for x in bis)
        if (
            bi1.direction == Direction.Up
            and bi1.low == min_low
            and bi2.high == max_high
            and bi2_start_dif < abs(zs_dif) * 0.5
            and bi1_dif > bi2_dif > zs_dif > 0
        ):
            return create_single_signal(k1=k1, k2=k2, k3=k3, v1="看空", v2=f"上涨{n}笔")

        if (
            bi1.direction == Direction.Down
            and bi1.high == max_high
            and bi2.low == min_low
            and bi2_start_dif > abs(zs_dif) * 0.5
            and 0 > zs_dif > bi2_dif > bi1_dif
        ):
            return create_single_signal(k1=k1, k2=k2, k3=k3, v1="看多", v2=f"下跌{n}笔")

    return create_single_signal(k1=k1, k2=k2, k3=k3, v1=v1)


def zdy_zs_space_V230421(c: CZSC, **kwargs):
    """中枢空间形态约束

    参数模板："{freq}_D{di}中枢空间_BS辅助V230421"

    **信号逻辑：**

    以顶背驰为例：要求【离开笔】距离中枢上沿的高度，大于等于【进入笔】距离中枢下沿的高度

    **信号列表：**

    - Signal('15分钟_D1中枢空间_BS辅助V230421_上涨_5笔_任意_0')
    - Signal('15分钟_D1中枢空间_BS辅助V230421_下跌_5笔_任意_0')
    - Signal('15分钟_D1中枢空间_BS辅助V230421_上涨_9笔_任意_0')
    - Signal('15分钟_D1中枢空间_BS辅助V230421_上涨_7笔_任意_0')
    - Signal('15分钟_D1中枢空间_BS辅助V230421_下跌_7笔_任意_0')
    - Signal('15分钟_D1中枢空间_BS辅助V230421_下跌_9笔_任意_0')

    :param c: 基础周期的 CZSC 对象
    :param kwargs: 其他参数
        - di: 倒数第 di 根 K 线
    :return: 信号字典
    """
    di = int(kwargs.get('di', 1))
    freq = c.freq.value
    k1, k2, k3 = f"{freq}_D{di}中枢空间_BS辅助V230421".split('_')
    v1 = '其他'
    if len(c.bi_list) < 7 or len(c.bars_ubi) > 7:
        return create_single_signal(k1=k1, k2=k2, k3=k3, v1=v1)

    for n in (9, 7, 5):
        bis = get_sub_elements(c.bi_list, di=di, n=n)
        if len(bis) != n:
            continue

        # 假定离开中枢的都是一笔
        zs = ZS(bis[1:-1])
        if not zs.is_valid:
            continue

        bi1, bi2 = bis[0], bis[-1]
        min_low = min(x.low for x in bis)
        max_high = max(x.high for x in bis)
        if (
            bi1.direction == Direction.Up
            and bi1.low == min_low
            and bi2.high == max_high
            and bi2.high - zs.zg >= zs.zd - bi1.low
        ):
            return create_single_signal(k1=k1, k2=k2, k3=k3, v1="上涨", v2=f"{n}笔")

        if (
            bi1.direction == Direction.Down
            and bi1.high == max_high
            and bi2.low == min_low
            and zs.zd - bi2.low >= bi1.high - zs.zg
        ):
            return create_single_signal(k1=k1, k2=k2, k3=k3, v1="下跌", v2=f"{n}笔")

    return create_single_signal(k1=k1, k2=k2, k3=k3, v1=v1)


def zdy_macd_dif_V230516(c: CZSC, **kwargs) -> OrderedDict:
    """MACD柱子与DIF的关系

    参数模板："{freq}_D{di}DIF走平_BS辅助V230516"

    **信号逻辑：**

    * DIF走平看多：1） DIF小于MACD柱子；2）DIF相比前一周期下跌小于阈值
    * DIF走平看空：1） DIF大于MACD柱子；2）DIF相比前一周期上涨小于阈值

    **信号列表：**

    - Signal('60分钟_D1DIF走平_BS辅助V230516_看空_红柱远离_任意_0')
    - Signal('60分钟_D1DIF走平_BS辅助V230516_看空_柱子否定_任意_0')
    - Signal('60分钟_D1DIF走平_BS辅助V230516_看多_绿柱远离_任意_0')
    - Signal('60分钟_D1DIF走平_BS辅助V230516_看多_柱子否定_任意_0')

    :param c: CZSC对象
    :param kwargs: 参数字典
    :return: 返回信号结果
    """
    di = int(kwargs.get('di', 1))
    freq = c.freq.value
    cache_key = update_macd_cache(c, fastperiod=12, slowperiod=26, signalperiod=9)
    k1, k2, k3 = f"{freq}_D{di}DIF走平_BS辅助V230516".split('_')
    v1, v2 = '其他', '其他'
    if len(c.bars_raw) < 12 + di:
        return create_single_signal(k1=k1, k2=k2, k3=k3, v1=v1)

    bars = get_sub_elements(c.bars_raw, di=di, n=10)
    # macd = bars[-1].cache[cache_key]['macd'] * 2
    dif = [x.cache[cache_key]['dif'] for x in bars]

    dif_th = sum([abs(x - y) for x, y in zip(dif, dif[1:])]) / len(dif) * 0.2
    if dif[-1] - dif[-2] > -dif_th:
        v1 = '看多'
        min_macd = min([x.cache[cache_key]['macd'] for x in bars])
        v2 = "绿柱远离" if dif[-1] < min_macd * 2.5 else "柱子否定"

    if dif[-1] - dif[-2] < dif_th:
        v1 = '看空'
        max_macd = max([x.cache[cache_key]['macd'] for x in bars])
        v2 = "红柱远离" if dif[-1] > max_macd * 2.5 else "柱子否定"
    return create_single_signal(k1=k1, k2=k2, k3=k3, v1=v1, v2=v2)


def zdy_macd_dif_V230517(c: CZSC, **kwargs) -> OrderedDict:
    """MACD三次开仓条件

    参数模板："{freq}_D{di}MACD开仓_BS辅助V230517"

    **信号逻辑：**

    以多头开仓为例：

    1. DIF 在零轴下方运行很长时间后首次升破零轴
    2. DIF 在零轴上方首次出现与 DEA 的飞吻
    3. DIF 在零轴上方首次出现与 MACD 金叉

    **信号列表：**

    - Signal('60分钟_D1MACD开仓_BS辅助V230517_看空_DIF破零轴_任意_0')
    - Signal('60分钟_D1MACD开仓_BS辅助V230517_看空_MACD飞吻_任意_0')
    - Signal('60分钟_D1MACD开仓_BS辅助V230517_看空_MACD死叉_任意_0')
    - Signal('60分钟_D1MACD开仓_BS辅助V230517_看多_DIF破零轴_任意_0')
    - Signal('60分钟_D1MACD开仓_BS辅助V230517_看多_MACD飞吻_任意_0')
    - Signal('60分钟_D1MACD开仓_BS辅助V230517_看多_MACD金叉_任意_0')

    :param c: CZSC对象
    :param kwargs: 参数字典
    :return: 返回信号结果
    """
    di = int(kwargs.get('di', 1))
    freq = c.freq.value
    cache_key = update_macd_cache(c, fastperiod=12, slowperiod=26, signalperiod=9)
    k1, k2, k3 = f"{freq}_D{di}MACD开仓_BS辅助V230517".split('_')
    if len(c.bars_raw) < 50:
        return create_single_signal(k1=k1, k2=k2, k3=k3, v1='其他')

    bars = get_sub_elements(c.bars_raw, di=di, n=20)
    macd = [x.cache[cache_key]['macd'] for x in bars]
    dif = [x.cache[cache_key]['dif'] for x in bars]

    if dif[-1] > 0:
        v1 = '看多'
        v2 = None
        if all([x < 0 for x in dif[:-1]]):
            v2 = 'DIF破零轴'

        if macd[-1] > 0 and macd[-2] < 0:
            v2 = 'MACD金叉'

        if macd[-5] > macd[-4] > macd[-3] > macd[-2] < macd[-1] and macd[-2] > 0:
            v2 = 'MACD飞吻'

        if v2:
            return create_single_signal(k1=k1, k2=k2, k3=k3, v1=v1, v2=v2)

    if dif[-1] < 0:
        v1 = '看空'
        v2 = None
        if all([x > 0 for x in dif[:-1]]):
            v2 = 'DIF破零轴'

        if macd[-1] < 0 and macd[-2] > 0:
            v2 = 'MACD死叉'

        if macd[-5] < macd[-4] < macd[-3] < macd[-2] > macd[-1] and macd[-2] < 0:
            v2 = 'MACD飞吻'

        if v2:
            return create_single_signal(k1=k1, k2=k2, k3=k3, v1=v1, v2=v2)

    return create_single_signal(k1=k1, k2=k2, k3=k3, v1='其他')


def zdy_macd_V230518(c: CZSC, **kwargs) -> OrderedDict:
    """MACD交叉次数

    参数模板："{freq}_D{di}MACD交叉N{n}_BS辅助V230518"

     **信号逻辑：**

    1. MACD 大于0，金叉；MACD 小于0，死叉
    2. 计算 MACD 连续大于0或者小于0的次数

     **信号列表：**

    - Signal('60分钟_D1MACD交叉N9_BS辅助V230518_死叉_第1次_任意_0')
    - Signal('60分钟_D1MACD交叉N9_BS辅助V230518_死叉_第2次_任意_0')
    - Signal('60分钟_D1MACD交叉N9_BS辅助V230518_死叉_第3次_任意_0')
    - Signal('60分钟_D1MACD交叉N9_BS辅助V230518_死叉_第4次_任意_0')
    - Signal('60分钟_D1MACD交叉N9_BS辅助V230518_死叉_第5次_任意_0')
    - Signal('60分钟_D1MACD交叉N9_BS辅助V230518_死叉_第6次_任意_0')
    - Signal('60分钟_D1MACD交叉N9_BS辅助V230518_死叉_第7次_任意_0')
    - Signal('60分钟_D1MACD交叉N9_BS辅助V230518_死叉_第8次_任意_0')
    - Signal('60分钟_D1MACD交叉N9_BS辅助V230518_死叉_第9次_任意_0')
    - Signal('60分钟_D1MACD交叉N9_BS辅助V230518_金叉_第1次_任意_0')
    - Signal('60分钟_D1MACD交叉N9_BS辅助V230518_金叉_第2次_任意_0')
    - Signal('60分钟_D1MACD交叉N9_BS辅助V230518_金叉_第3次_任意_0')
    - Signal('60分钟_D1MACD交叉N9_BS辅助V230518_金叉_第4次_任意_0')
    - Signal('60分钟_D1MACD交叉N9_BS辅助V230518_金叉_第5次_任意_0')
    - Signal('60分钟_D1MACD交叉N9_BS辅助V230518_金叉_第6次_任意_0')
    - Signal('60分钟_D1MACD交叉N9_BS辅助V230518_金叉_第7次_任意_0')
    - Signal('60分钟_D1MACD交叉N9_BS辅助V230518_金叉_第8次_任意_0')
    - Signal('60分钟_D1MACD交叉N9_BS辅助V230518_金叉_第9次_任意_0')

    :param c: CZSC对象
    :param kwargs: 参数字典
     :return: 返回信号结果
    """
    di = int(kwargs.get('di', 1))
    n = int(kwargs.get('n', 9))
    freq = c.freq.value
    cache_key = update_macd_cache(c, fastperiod=12, slowperiod=26, signalperiod=9)
    k1, k2, k3 = f"{freq}_D{di}MACD交叉N{n}_BS辅助V230518".split('_')
    if len(c.bars_raw) < 50:
        return create_single_signal(k1=k1, k2=k2, k3=k3, v1='其他')

    bars = get_sub_elements(c.bars_raw, di=di, n=n + 1)
    macd = [x.cache[cache_key]['macd'] for x in bars]
    v1 = '金叉' if macd[-1] > 0 else '死叉'

    count = 0
    for m in macd[::-1]:
        if (m > 0 and macd[-1] > 0) or (m < 0 and macd[-1] < 0):
            count += 1
        else:
            break

    if count == n + 1:
        return create_single_signal(k1=k1, k2=k2, k3=k3, v1=v1, v2="超计数范围")

    return create_single_signal(k1=k1, k2=k2, k3=k3, v1=v1, v2=f"第{count}次")


def zdy_macd_V230519(c: CZSC, **kwargs) -> OrderedDict:
    """MACD连续缩柱

    参数模板："{freq}_D{di}N{n}MACD缩柱_BS辅助V230519"

    **信号逻辑：**

    1. 取 N 跟 K 线，如果 N 跟 K 线的 MACD 都在零轴上方，且 N 跟 K 线的 MACD 都比 N-1 跟 K-1 线的 MACD 小，则认为是多头连续缩柱
    2. 反之为空头连续缩柱

    **信号列表：**

    - Signal('60分钟_D1N3MACD缩柱_BS辅助V230519_多头连续缩柱_任意_任意_0')
    - Signal('60分钟_D1N3MACD缩柱_BS辅助V230519_空头连续缩柱_任意_任意_0')

    :param c: CZSC对象
    :param kwargs: 参数字典
    :return: 返回信号结果
    """
    di = int(kwargs.get('di', 1))
    n = int(kwargs.get('n', 3))
    freq = c.freq.value
    cache_key = update_macd_cache(c, fastperiod=12, slowperiod=26, signalperiod=9)
    k1, k2, k3 = f"{freq}_D{di}N{n}MACD缩柱_BS辅助V230519".split('_')
    v1 = '其他'
    if len(c.bars_raw) < 50:
        return create_single_signal(k1=k1, k2=k2, k3=k3, v1=v1)

    bars = get_sub_elements(c.bars_raw, di=di, n=n)
    macd = [x.cache[cache_key]['macd'] for x in bars]

    if all([x > 0 for x in macd]) and all([macd[i] < macd[i - 1] for i in range(1, n)]):
        v1 = '多头连续缩柱'

    if all([x < 0 for x in macd]) and all([macd[i] > macd[i - 1] for i in range(1, n)]):
        v1 = '空头连续缩柱'

    return create_single_signal(k1=k1, k2=k2, k3=k3, v1=v1)


def zdy_macd_dif_iqr_V230521(c: CZSC, **kwargs) -> OrderedDict:
    """MACD柱子与DIF的关系

    参数模板："{freq}_D{di}DIF走平IQR_BS辅助V230521"

    **信号逻辑：**

    * DIF走平看多：1） DIF小于MACD柱子；2）最近3个周期DIF相比前一周期变化在四分位距内
    * DIF走平看空：1） DIF大于MACD柱子；2）最近3个周期DIF相比前一周期变化在四分位距内

    **信号列表：**

    - Signal('60分钟_D1DIF走平IQR_BS辅助V230521_看空_红柱远离_任意_0')
    - Signal('60分钟_D1DIF走平IQR_BS辅助V230521_看空_柱子否定_任意_0')
    - Signal('60分钟_D1DIF走平IQR_BS辅助V230521_看多_绿柱远离_任意_0')
    - Signal('60分钟_D1DIF走平IQR_BS辅助V230521_看多_柱子否定_任意_0')

    :param c: CZSC对象
    :param kwargs: 参数字典
    :return: 返回信号结果
    """
    di = int(kwargs.get('di', 1))
    freq = c.freq.value
    cache_key = update_macd_cache(c, fastperiod=12, slowperiod=26, signalperiod=9)
    k1, k2, k3 = f"{freq}_D{di}DIF走平IQR_BS辅助V230521".split('_')
    v1, v2 = '其他', '其他'
    if len(c.bars_raw) < 50:
        return create_single_signal(k1=k1, k2=k2, k3=k3, v1=v1)

    bars = get_sub_elements(c.bars_raw, di=di, n=100)
    macd = bars[-1].cache[cache_key]['macd'] * 2
    dif = [x.cache[cache_key]['dif'] for x in bars]

    # 计算四分位距
    Q3, Q1 = np.percentile(dif, [75, 25])
    IQR = Q3 - Q1

    if max(dif[-3:]) - min(dif[-3:]) < IQR and macd < 0:
        v1 = '看多'
        v2 = "绿柱远离" if dif[-1] < macd else "柱子否定"

    if max(dif[-3:]) - min(dif[-3:]) < IQR and macd > 0:
        v1 = '看空'
        v2 = "红柱远离" if dif[-1] > macd else "柱子否定"
    return create_single_signal(k1=k1, k2=k2, k3=k3, v1=v1, v2=v2)


def zdy_macd_V230527(c: CZSC, **kwargs) -> OrderedDict:
    """DIF/DEA/MACD 远离零轴辅助判断买卖点

    参数模板："{freq}_{key}远离W{w}N{n}T{t}_BS辅助V230527"

    **信号逻辑：**

    1. 获取最近 w 根K线，计算 DIF/DEA/MACD，计算绝对值的中位数和标准差
    2. 如果最新的 n 个值的最大绝对值大于中位数 + t / 10 * 标准差，则认为远离零轴

    **信号列表：**

    - Signal('60分钟_DIF远离W100N10T10_BS辅助V230527_多头远离_任意_任意_0')
    - Signal('60分钟_DIF远离W100N10T10_BS辅助V230527_空头远离_任意_任意_0')

    :param c: CZSC对象
    :return: 信号识别结果
    """
    cache_key = update_macd_cache(c, fastperiod=12, slowperiod=26, signalperiod=9)
    n = int(kwargs.get("n", 10))
    w = int(kwargs.get("w", 100))
    t = int(kwargs.get("t", 20))  # 远离零轴的阈值，越大越远离零轴，比较的基准是窗口内的中位数和标准差；20表示2倍
    key = kwargs.get("key", "dif").upper()
    assert key in ["DIF", "DEA", "MACD"]
    freq = c.freq.value
    k1, k2, k3 = f"{freq}_{key}远离W{w}N{n}T{t}_BS辅助V230527".split("_")
    v1 = '其他'
    if len(c.bi_list) < 3:
        return create_single_signal(k1=k1, k2=k2, k3=k3, v1=v1)

    bars = get_sub_elements(c.bars_raw, di=1, n=w)
    factors = [x.cache[cache_key][key.lower()] for x in bars]
    median = np.median(np.abs(factors))
    std = np.std(np.abs(factors))

    last_n_factors = factors[-n:]
    max_abs_factor = max(last_n_factors, key=abs)
    if abs(max_abs_factor) > median + t / 10 * std:
        v1 = f"{'多头' if max_abs_factor > 0 else '空头'}远离"
        return create_single_signal(k1=k1, k2=k2, k3=k3, v1=v1)

    return create_single_signal(k1=k1, k2=k2, k3=k3, v1=v1)


def zdy_dif_V230527(c: CZSC, **kwargs) -> OrderedDict:
    """DIF远离零轴辅助判断买卖点

    参数模板："{freq}_N{n}T{t}_DIF远离V230527"

    **信号逻辑：**

    以多头远离为例：回溯N根K线，找到DIF最大值，然后绿柱回溯10跟找到MACD最大值，然后看最大值之间的距离判断原理。

    **信号列表：**

    - Signal('5分钟_N10T30_DIF远离V230527_空头远离_任意_任意_0')
    - Signal('5分钟_N10T30_DIF远离V230527_多头远离_任意_任意_0')

    :param c: CZSC对象
    :return: 信号识别结果
    """
    cache_key = update_macd_cache(c, fastperiod=12, slowperiod=26, signalperiod=9)
    n = int(kwargs.get("n", 10))
    t = int(kwargs.get("t", 30))  # 远离零轴的阈值，越大越远离零轴；30表示3倍
    freq = c.freq.value
    k1, k2, k3 = f"{freq}_N{n}T{t}_DIF远离V230527".split("_")
    v1 = '其他'
    if len(c.bars_raw) < 30 + n * 8:
        return create_single_signal(k1=k1, k2=k2, k3=k3, v1=v1)

    bars = get_sub_elements(c.bars_raw, di=1, n=n * 8)
    max_abs_dif_bar = max(bars[-n:], key=lambda x: abs(x.cache[cache_key]['dif']))
    max_abs_dif = max_abs_dif_bar.cache[cache_key]['dif']

    if max_abs_dif > 0:
        macd_seq = [x.cache[cache_key]['macd'] for x in bars if x.cache[cache_key]['macd'] > 0]
        if len(macd_seq) > n and abs(max_abs_dif) > max(macd_seq) * t / 10:
            v1 = '多头远离'
    elif max_abs_dif < 0:
        macd_seq = [abs(x.cache[cache_key]['macd']) for x in bars if x.cache[cache_key]['macd'] < 0]
        if len(macd_seq) > n and abs(max_abs_dif) > max(macd_seq) * t / 10:
            v1 = '空头远离'

    return create_single_signal(k1=k1, k2=k2, k3=k3, v1=v1)


def zdy_dif_V230528(c: CZSC, **kwargs) -> OrderedDict:
    """DIF远离零轴辅助判断买卖点

    参数模板："{freq}_N{n}T{t}_DIF远离V230528"

    **信号逻辑：**

    以多头远离为例：回溯1000根K线，找到DIF的所有峰谷值，如最近一个是峰值，且大于所有峰值的 T% 分位数。

    **信号列表：**

    - Signal('5分钟_N20T70_DIF远离V230528_空头远离_任意_任意_0')
    - Signal('5分钟_N20T70_DIF远离V230528_多头远离_任意_任意_0')

    :param c: CZSC对象
    :return: 信号识别结果
    """
    cache_key = update_macd_cache(c, fastperiod=12, slowperiod=26, signalperiod=9)
    n = int(kwargs.get("n", 20))  # 峰谷值的数量最小值
    t = int(kwargs.get("t", 80))  # 峰谷排序的分位数
    freq = c.freq.value
    k1, k2, k3 = f"{freq}_N{n}T{t}_DIF远离V230528".split("_")
    v1 = '其他'

    def _find_peaks_valleys(data):
        """在给定的一维数据中找到所有的峰值和谷值。

        :param data: 输入的一维数据，应为一个列表或者类似列表的数据结构
        :return: 峰值和谷值的索引和对应的值
        """
        peaks = {}
        valleys = {}

        # 检查数据是否至少有5个元素
        if len(data) < 5:
            return {"peaks": peaks, "valleys": valleys}

        # 遍历数据，找到所有的峰值和谷值
        for i in range(2, len(data) - 2):
            # 如果当前元素比它的左右两侧的元素都大，则它是一个峰值
            if data[i - 2] < data[i - 1] < data[i] > data[i + 1] > data[i + 2]:
                peaks[i] = data[i]

            # 如果当前元素比它的左右两侧的元素都小，则它是一个谷值
            if data[i - 2] > data[i - 1] > data[i] < data[i + 1] < data[i + 2]:
                valleys[i] = data[i]

        return peaks, valleys

    dif_values = [x.cache[cache_key]['dif'] for x in c.bars_raw[-1000:]]
    peaks, valleys = _find_peaks_valleys(dif_values)

    if len(peaks) < n or len(valleys) < n:
        return create_single_signal(k1=k1, k2=k2, k3=k3, v1=v1)

    peaks_n = np.percentile(list(peaks.values()), t)
    valleys_n = np.percentile(list(valleys.values()), 100 - t)

    if max(peaks.keys()) > max(valleys.keys()) and peaks[max(peaks.keys())] > peaks_n and dif_values[-1] > 0:
        v1 = '多头远离'

    if max(valleys.keys()) > max(peaks.keys()) and valleys[max(valleys.keys())] < valleys_n and dif_values[-1] < 0:
        v1 = '空头远离'

    return create_single_signal(k1=k1, k2=k2, k3=k3, v1=v1)
